/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: af_l2cap.c,v 1.26 2002/11/27 16:26:59 kds Exp $

   AF_L2CAP - L2CAP Address family for socket interface

   Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
                Imre Deak <ext-imre.deak@nokia.com>
*/		

/* The following prevents "kernel_version" from being set in this file. */
#define __NO_VERSION__

#include <linux/config.h>
#include <linux/version.h>

/* Module related headers, non-module drivers should not include */
#include <linux/module.h>
#include <linux/init.h>

/* Standard driver includes */
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/slab.h>
#include <linux/timer.h>
#include <linux/spinlock.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/ioctl.h>
#include <linux/proc_fs.h>
#include <linux/notifier.h>
#include <asm/uaccess.h>
#include <asm/bitops.h>

#include <linux/in.h>
#include <linux/skbuff.h>
#include <linux/socket.h>
#include <linux/net.h>
#include <net/sock.h>

/* Local Includes */
#define FILEBIT	DBAFL2CAP

#include <affix/bluetooth.h>
#include <affix/btdebug.h>
#include <affix/hci.h>
#include <affix/l2cap.h>

typedef struct l2cap_sock	l2cap_sock;

struct l2cap_sock{
	struct list_head	q;

	struct sock	*sk;
	BD_ADDR		bda;
	__u16		psm, sport;
	l2cap_ch	*ch;
	__u8		hdrlen;

	/* l2cap options */
	int		mtu, mru;
};

static inline l2cap_sock *lpf_get(struct sock *sk)
{
	return (l2cap_sock*)sk->prot;
}


/* LPF driver variables */
btlist_head_t			l2cap_socks;
spinlock_t			lpf_spinlock = SPIN_LOCK_UNLOCKED;
extern l2cap_proto_ops 		lpf_proto;

/*
 * debuging stuff
 */
static long sock_count = 0;


/***************************    Connection Management   ********************************/

struct sock *lpf_lookup_sock(hci_struct *hci, __u16 psm)
{
	l2cap_sock	*lsk, *zero = NULL;
	struct sock	*sk = NULL;
	
	btl_read_lock(&l2cap_socks);
	btl_for_each (lsk, l2cap_socks) {
		if (lsk->sk->state == CON_LISTEN && lsk->psm == psm) {
			if (!zero && bda_zero(&lsk->bda))
				zero = lsk;
			if (bda_equal(&hci->bda, &lsk->bda))
				break;
		}
	}
	if (!lsk)
		lsk = zero;
	if (lsk) {
		sock_hold(lsk->sk);
		sk = lsk->sk;
	}
	btl_read_unlock(&l2cap_socks);

	return sk;
}

l2cap_sock *lpf_alloc(int priority)
{
	l2cap_sock	*lsk;

	lsk = (l2cap_sock*)kmalloc(sizeof(l2cap_sock), priority);
	if (lsk == NULL)
		return NULL;
	memset(lsk, 0, sizeof(l2cap_sock));

	return lsk;
}

void lpf_destruct(struct sock *sk)
{
	l2cap_sock	*lsk = lpf_get(sk);

	DBFENTER;
	
	kfree(lsk);
	sock_count--;	/* XXX */

	DBFEXIT;
}

void lpf_destroy(struct sock *sk)
{
	l2cap_sock	*lsk = lpf_get(sk);
	struct sk_buff	*skb;
	
	DBFENTER;

	btl_unlink(&l2cap_socks, lsk);

	if (lsk->ch) {
		lsk->ch->priv = NULL;
		l2ca_put(xchg(&lsk->ch, NULL));
	}
	
	if (sk->pair) {
		sock_put(sk->pair);
		sk->pair = NULL;
	}

	/* remove all packets and check if threre is a connection in the queue */
	while((skb = skb_dequeue(&sk->receive_queue)) != NULL) {
		if (skb->sk != sk)	/* A pending connection */
			lpf_destroy(skb->sk);
		kfree_skb(skb);
	}

	sock_put(sk);

	DBFEXIT;
}

struct sock *lpf_create(int family, int priority, int zero_it)
{
	struct sock	*sk;
	l2cap_sock	*lsk;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	sk  = sk_alloc(family, priority, zero_it);
#else
	sk  = sk_alloc(family, priority, zero_it, NULL);
#endif
	if (sk == NULL)
		return NULL;

	lsk = lpf_alloc(priority);
	if (lsk == NULL) {
		sk_free(sk);
		return NULL;
	}

	lsk->sk = sk;
	sk->prot = (void*)lsk;

	// set some options
	lsk->mru = 0x8000;	/* FIXME */

	btl_add_tail(&l2cap_socks, lsk);

	sock_count++;	/* XXX */

	return sk;
}

void lpf_init_data(struct socket *sock, struct sock *sk)
{
	DBFENTER;
	
	sock_init_data(sock, sk);
	sk->destruct = lpf_destruct;
	
	DBFEXIT;
}

int lpf_wait_for_state(struct sock *sk, __u8 state, int noblock)
{
	int	err = 0;
	long	timeo;
	DECLARE_WAITQUEUE(wait, current);

	DBFENTER;

	timeo = sock_sndtimeo(sk, noblock);

	add_wait_queue(sk->sleep, &wait);
	set_current_state(TASK_INTERRUPTIBLE);

	while (sk->state != state) {
		if (!timeo) {
			err = -EAGAIN;
			break;
		}
		if ((err = sock_error(sk))) 
			break;
		if (signal_pending(current)) {
			err = -ERESTARTSYS;
			break;
		}
		timeo = schedule_timeout(timeo);
		set_current_state(TASK_INTERRUPTIBLE);
	}

	set_current_state(TASK_RUNNING);
	remove_wait_queue(sk->sleep, &wait);

	DBFEXIT;
	return err;
}

int lpf_connect_req(struct sock *sk, struct sockaddr_affix *addr)
{
	l2cap_sock 	*lsk = lpf_get(sk);
	l2cap_ch	*ch;
	int		err;
	hci_struct	*hci;

	DBFENTER;

	if (!bda_zero(&addr->local)) {
		hci = hci_lookup_bda(&addr->local);	/* find device */
		if (hci == NULL)
			return -ENODEV;
	} else
		hci = NULL;

	lsk->bda = addr->bda;
	lsk->psm = addr->port;
	
	ch = l2ca_create(lsk->psm, &lpf_proto);
	if (ch == NULL) {
		if (hci)
			hci_put(hci);
		err = -ENOMEM;
		goto exit;
	}

	lsk->ch = ch;
	ch->priv = sk;
	ch->hci = hci;

	l2ca_setmtu(lsk->ch, lsk->mru);
	err = l2ca_connect_req(ch, &lsk->bda);

	lsk->hdrlen = ch->hdrlen;
 exit:
	DBFEXIT;
	return err;
}

void lpf_shutdown(struct sock *sk, int reason)
{
	if (reason) {
		switch (reason) {
			case 0xEEEE:
				sk->err = ETIMEDOUT;
				break;
			case CON_RSP_PSM:
				sk->err = ECONNREFUSED;
				break;
			default:
				sk->err = ECONNRESET;
		}
	} 

        sk->shutdown |= RCV_SHUTDOWN;
	sk->state = CON_CLOSED;
	if (!sk->dead) {
                sk->dead = 1;
		sk->state_change(sk);
        }
}

/* 
   LPF protocol subsystem.  Client of L2CAP. Callback functions
*/

int lpf_data_ind(l2cap_ch *ch, struct sk_buff *skb)
{
	struct sock	*sk = (struct sock*)ch->priv;

	DBFENTER;

	// FIXME: Check state of the socket

	affix_sock_rx(sk, skb);

	DBFEXIT;
	return 0;
}

struct sock *lpf_dup(struct sock *osk)
{
	struct sock	*sk;

	sk = lpf_create(PF_AFFIX, GFP_ATOMIC, 1);
	if (sk == NULL)
		return NULL;

	lpf_init_data(NULL, sk);

	sk->pair = osk;

	sk->protocol = osk->protocol;
	sk->type     = osk->type;
	sk->destruct = osk->destruct;
	sk->sleep    = osk->sleep;

	sk->socket   = osk->socket;
	sk->priority = osk->priority;

	return sk;
}
/*
 * ch hold here. need to put it
 */
int lpf_connect_ind(l2cap_ch *ch)
{
	struct sock	*sk, *pair;
	l2cap_sock	*lsk;

	DBFENTER;

	pair = lpf_lookup_sock(ch->hci, ch->psm);
	if (pair == NULL) {
		l2ca_connect_rsp(ch, CON_RSP_PSM, 0);
		return -1;
	}
	if (pair->state != CON_LISTEN || pair->ack_backlog >= pair->max_ack_backlog) {
		l2ca_connect_rsp(ch, CON_RSP_RESOURCE, 0);
		sock_put(pair);
		return -1;
	}

	sk = lpf_dup(pair);
	if (sk == NULL) {
		l2ca_connect_rsp(ch, CON_RSP_RESOURCE, 0);
		sock_put(pair);
		return -1;
	}

	pair->ack_backlog++;

	lsk = lpf_get(sk);
	l2ca_hold(ch);		// hold it
	lsk->ch = ch;
	lsk->hdrlen = ch->hdrlen;
	ch->priv = sk;
	lsk->bda = ch->bda;
	l2ca_setmtu(lsk->ch, lsk->mru);
	l2ca_connect_rsp(ch, CON_RSP_SUCCESS, 0);
	
	DBFEXIT;
	return 0;
}

int lpf_connect_cfm(l2cap_ch *ch, __u16 result, __u16 status)
{
	struct sock	*sk = (struct sock*)ch->priv;
	l2cap_sock	*lsk = lpf_get(sk);
	struct sk_buff	*skb;
	
	DBFENTER;

	/* actually  it needed to analyaze result */
	if (result) {
		if (sk->pair)
			lpf_destroy(sk);
		else
			lpf_shutdown(sk, result);
		return 0;
	}

	lsk->hdrlen = ch->hdrlen;
	sk->state = CON_ESTABLISHED;
	lsk->mtu = l2ca_getmtu(ch);

	if (sk->pair){ /* incomming connection */
		if (!sk->pair->dead) {
			skb = alloc_skb(0, GFP_ATOMIC); 
			skb_set_owner_r(skb, sk);
			skb_queue_tail(&sk->pair->receive_queue, skb);
			sk->pair->data_ready(sk->pair, skb->len);
		} else
			lpf_destroy(sk);	/* not owned by "socket" */
	} else if (!sk->dead)
		sk->state_change(sk);
	
	DBFEXIT;
	return 0;
}


int lpf_config_ind(l2cap_ch *ch)
{
	struct sock	*sk = (struct sock*)ch->priv;
	l2cap_sock	*lsk = lpf_get(sk);
	__u16		result = 0;

	DBFENTER;

	l2ca_setmtu(lsk->ch, lsk->mru);
	result = l2ca_config_rsp(ch, CFG_RSP_SUCCESS);

	DBFEXIT;
	return result;
}

int lpf_config_cfm(l2cap_ch *ch, __u16 result)
{
	struct sock	*sk = (struct sock*)ch->priv;
	l2cap_sock	*lsk = lpf_get(sk);

	DBFENTER;

	if (result)
		lpf_shutdown(sk, CON_RSP_PSM);	// like in disconnect_ind

	lsk->mtu = l2ca_getmtu(ch);
	
	DBFEXIT;
	return 0;
}

int lpf_disconnect_ind(l2cap_ch *ch)
{
	struct sock	*sk = (struct sock*)ch->priv;

	DBFENTER;

	lpf_shutdown(sk, 0);

	DBFEXIT;
	return 0;
}

/* Callback functions for L2CAP layer */
l2cap_proto_ops lpf_proto = {
	owner:		THIS_MODULE,
	data_ind:	lpf_data_ind,
	connect_ind:	lpf_connect_ind,
	connect_cfm:	lpf_connect_cfm,
	config_ind:	lpf_config_ind,
	config_cfm:	lpf_config_cfm,
	disconnect_ind:	lpf_disconnect_ind
};


/******************************  L2CAP PROTOCOL FAMILY SUBSYSTEM  **************************/

static inline void lpf_wait_on_queue(struct sock *sk)
{
	DECLARE_WAITQUEUE(wait, current);

	DBFENTER;

	add_wait_queue(sk->sleep, &wait);
	set_current_state(TASK_INTERRUPTIBLE);

	if (skb_peek(&sk->receive_queue) == NULL)
		schedule();

	set_current_state(TASK_RUNNING);
	remove_wait_queue(sk->sleep, &wait);

	DBFEXIT;
}

int lpf_release(struct socket *sock)
{
	struct sock	*sk = sock->sk;
	l2cap_sock	*lsk;

	DBFENTER;

	if (sk == NULL)
		return 0;

        sock->sk   = NULL;      
        sk->socket = NULL;      /* Not used, but we should do this. */

	lsk = lpf_get(sk);
	
	if (lsk->sport) {
		int		count = 0;
		l2cap_sock	*_lsk;
		/* count registered sockets */
		btl_read_lock(&l2cap_socks);
		btl_for_each (_lsk, l2cap_socks) {
			if (lsk->sport == _lsk->sport)
				count++;
		}
		btl_read_unlock(&l2cap_socks);
		if (count == 1)	/* we only use it */
			l2ca_unregister_protocol(lsk->sport);
	}

	sk->state = CON_CLOSED;
        sk->err = 0;
	sk->shutdown |= SEND_SHUTDOWN;
	sk->dead = 1;

	lpf_destroy(sk);

	MOD_DEC_USE_COUNT;

	DBFEXIT;
	return 0;
}

int lpf_bind(struct socket *sock, struct sockaddr *addr, int addr_len)
{
	struct sock		*sk = sock->sk;
	l2cap_sock		*lsk;
	struct sockaddr_affix	*saddr = (void*)addr;
	int			err = -EBUSY, count = 0;

	DBFENTER;

	/* find if already registered */
	btl_read_lock(&l2cap_socks);
	btl_for_each (lsk, l2cap_socks) {
		if (lsk->sport == saddr->port) {
			count++;
			if (bda_equal(&lsk->bda, &saddr->bda))
				break;
		}
	}
	if (!lsk) {
		err = 0;
		if (!count)
			err = l2ca_register_protocol(saddr->port, &lpf_proto);
		if (!err) {
			lsk = lpf_get(sk);
			lsk->sport = saddr->port;
			lsk->psm = saddr->port;
			lsk->bda = saddr->bda;
		}
	}
	btl_read_unlock(&l2cap_socks);

	DBFEXIT;
	return err;
}

/*
    Connect to remote device
    Wait until connection established
*/
int lpf_connect(struct socket *sock, struct sockaddr *addr, int alen, int flags)
{
	struct sock		*sk = sock->sk;
	struct sockaddr_affix	*saddr = (void*)addr;
	int			err = 0;

	DBFENTER;

	/* deal with restarts */
	if (sk->state == CON_ESTABLISHED && sock->state == SS_CONNECTING) {
		sock->state = SS_CONNECTED;
		return 0;
	}
		
	if (sk->state == CON_CLOSED && sock->state == SS_CONNECTING) {
		sock->state = SS_UNCONNECTED;
		return -ECONNREFUSED;
	}
	
	if (sk->state == CON_ESTABLISHED)
		return -EISCONN;      /* No reconnect on a seqpacket socket */
	
	/* Move to connecting socket, start sending Connect Requests */
	sock->state = SS_CONNECTING;
	sk->state   = CON_CONNECTING;

	err = lpf_connect_req(sk, saddr);
	if (err < 0)
		goto exit;

	err = lpf_wait_for_state(sk, CON_ESTABLISHED, flags & O_NONBLOCK);
	if (err)
		goto exit;

	if (sk->state == CON_ESTABLISHED) {
		sock->state = SS_CONNECTED;
	}

 exit:
	DBFEXIT;
	return err;
}

int lpf_ioctl_config(struct socket *sock)
{
	struct sock	*sk = sock->sk;
	l2cap_sock	*lsk = lpf_get(sk);
	int		err = 0;

	DBFENTER;

	if (sk->state == CON_ESTABLISHED && sock->state == SS_CONNECTING) {
		sock->state = SS_CONNECTED;
		return 0;
	}

	/* Move to connecting socket, start sending Connect Requests */
	sock->state = SS_CONNECTING;
	sk->state   = CON_CONNECTING;

	l2ca_setmtu(lsk->ch, lsk->mru);
	err = l2ca_config_req(lsk->ch);
	if (err < 0)
		goto exit;

	err = lpf_wait_for_state(sk, CON_ESTABLISHED, sock->file->f_flags & O_NONBLOCK);

 exit:
	DBFEXIT;
	return err;
}

int lpf_ioctl_flush(struct socket *sock)
{
	struct sock	*sk = sock->sk;
	int		err = 0;
	DECLARE_WAITQUEUE(wait, current);

	DBFENTER;

	add_wait_queue(sk->sleep, &wait);
	set_current_state(TASK_INTERRUPTIBLE);

	while (atomic_read(&sk->wmem_alloc) != 0) {
		if ((err = sock_error(sk))) 
			break;
		if (signal_pending(current)) {
			err = -ERESTARTSYS;
			break;
		}
		schedule();
		set_current_state(TASK_INTERRUPTIBLE);
	}

	set_current_state(TASK_RUNNING);
	remove_wait_queue(sk->sleep, &wait);

	DBFEXIT;
	return err;
}


int lpf_ioctl(struct socket *sock, unsigned int cmd,  unsigned long arg)
{
	DBFENTER;

	switch (cmd) {
		case SIOCL2CAP_CONFIG:
			return lpf_ioctl_config(sock);
		
		case SIOCL2CAP_FLUSH:
			return lpf_ioctl_flush(sock);

		default:
			return -EINVAL;
	}

	DBFEXIT;
	return 0;
}

/*
   register protocol and allow clients to establish connection
*/
int lpf_listen(struct socket *sock, int backlog)
{
	struct sock	*sk = sock->sk;
	l2cap_sock	*lsk = lpf_get(sk);

	DBFENTER;

	if (sock->state != SS_UNCONNECTED)
		return -EINVAL;

	if (lsk->sport == 0)
		return -EOPNOTSUPP;	

	if (sk->state != CON_LISTEN) {
		sk->max_ack_backlog = backlog;
		sk->state = CON_LISTEN;
		sock->flags |= __SO_ACCEPTCON;
	}
	
	DBFEXIT;
	return 0;
}


/*
   Accept connection from remote device
   Wait until ready connection will become available
*/

int lpf_accept(struct socket *sock, struct socket *newsock, int flags)
{
	struct sock	*sk = sock->sk, *newsk;
	struct sk_buff	*skb;
	int		err = -EINVAL;

	DBFENTER;

	if (sock->state != SS_UNCONNECTED || !(sock->flags & __SO_ACCEPTCON))
		goto exit;
	
	while( (skb = skb_dequeue(&sk->receive_queue)) == NULL) {
		if (flags & O_NONBLOCK)
			return -EWOULDBLOCK;

		lpf_wait_on_queue(sk);
		if (signal_pending(current)) 
			return -ERESTARTSYS;
	}

	if (newsock->sk) {
		sk_free(newsock->sk);
		newsock->sk = NULL;
	}

	newsk = skb->sk;
	sock_put(newsk->pair);		/* put pair */
	newsk->pair = NULL;
	sock_graft(newsk, newsock);
	newsock->state = SS_CONNECTED;
	sk->ack_backlog--;

	kfree_skb(skb);
	MOD_INC_USE_COUNT;
	err = 0;
 exit:			
	DBFEXIT;
	return err;
}

int lpf_getname(struct socket *sock, struct sockaddr *uaddr, int *alen, int peer)
{
	struct sock		*sk = sock->sk;
	struct sockaddr_affix	*addr = (struct sockaddr_affix*)uaddr;
	l2cap_sock		*lsk = lpf_get(sk);

	DBFENTER;

	addr->family = AF_AFFIX;
	addr->bda = lsk->bda;
	addr->port = lsk->psm;

	*alen = sizeof(struct sockaddr_affix);

	DBFEXIT;
	return 0;

}

int lpf_setsockopt(struct socket *sock, int level, int optname, char *optval, int optlen)
{
	l2cap_sock	*lsk = lpf_get(sock->sk);

	DBFENTER;
	switch (optname) {
		case BTSO_MTU:
			if (get_user(lsk->mru, (int*)optval))
				return -EFAULT;
			break;

		default:
			return -EINVAL;
	}
	DBFEXIT;
	return 0;
}

int lpf_getsockopt(struct socket *sock, int level, int optname, char *optval, int *optlen)
{
	l2cap_sock	*lsk = lpf_get(sock->sk);

	DBFENTER;
	switch (optname) {
		case BTSO_MTU:
			if (sock->sk->state != CON_ESTABLISHED)
				return -ENOTCONN;
			if (put_user(lsk->mtu, (int*)optval))
				return -EFAULT;
			if (put_user(sizeof(int), optlen))
				return -EFAULT;
			break;

		default:
			return -EINVAL;
	}
	DBFEXIT;
	return 0;
}


int lpf_getcmsg(struct msghdr *msg)
{
	DBPRT(DBXMIT, "control: %p, len: %d\n", msg->msg_control, msg->msg_controllen);
	if (msg->msg_control && msg->msg_controllen) {
		struct cmsghdr	*cmsg;
		for(cmsg = CMSG_FIRSTHDR(msg); cmsg; cmsg = CMSG_NXTHDR(msg, cmsg)) {
			if (cmsg->cmsg_level != SOL_AFFIX)
				continue;
			return cmsg->cmsg_type;
		}
	}
	return -1;
}

/*
    size - size of the iovec array, size of the message
*/
int lpf_sendmsg(struct socket *sock, struct msghdr *msg, int size, struct scm_cookie *scm)
{
	struct sock		*sk = sock->sk;
	l2cap_sock		*lsk;
	int			err = -ENOTCONN;
	struct sk_buff		*skb;
	int			hdrlen, ctype;

	DBFENTER;

	if (msg->msg_flags & ~MSG_DONTWAIT)
		return -EINVAL;

	if (sock->state != SS_CONNECTED) 
		return -ENOTCONN;

	if (sk->shutdown & SHUTDOWN_MASK) {
		if (!(msg->msg_flags & MSG_NOSIGNAL))
			send_sig(SIGPIPE, current, 0);
		return -EPIPE;
	}

	if (sk->state != CON_ESTABLISHED)
		return -ENOTCONN;

	lsk = lpf_get(sk);
	hdrlen = lsk->hdrlen;
	size = btmin(lsk->mtu, size);

	switch ((ctype = lpf_getcmsg(msg))) {
		case L2CAP_PING:
		case L2CAP_SINGLEPING:
			hdrlen += 4;	/* for command */	
			break;
	}
	
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 4)
	skb = sock_alloc_send_skb(sk, size + hdrlen, 0,
				  msg->msg_flags & MSG_DONTWAIT, &err);
#else
	skb = sock_alloc_send_skb(sk, size + hdrlen,
				  msg->msg_flags & MSG_DONTWAIT, &err);
#endif
	if (skb == NULL)
		return err;

	skb_reserve(skb, hdrlen);
	
	err = memcpy_fromiovec(skb_put(skb, size), msg->msg_iov, size);
	if (err)
		goto exit;

	DBPRTPKT(DBXMIT, skb->data, skb->len);

	switch (ctype) {
		case L2CAP_PING:
			DBPRT(DBXMIT, "Pinging... %d bytes\n", size);
			err = l2ca_ping(lsk->ch, skb);
			break;

		case L2CAP_SINGLEPING:
			DBPRT(DBXMIT, "Pinging... %d bytes\n", size);
			err = l2ca_singleping(lsk->ch, skb);
			break;

		default:
			/* normal data */
			err = l2ca_send_data(lsk->ch, skb);
	}

	if (err)
		goto exit;

	DBFEXIT;
	return size;
exit:
	kfree_skb(skb);
	return err;
}

int lpf_recvmsg(struct socket *sock, struct msghdr *msg, int size, int flags, struct scm_cookie *scm)
{
	struct sock		*sk = sock->sk;
	int			err = -ENOTCONN;
	struct sk_buff		*skb;
	int			copied;

	DBFENTER;

	if (sock->flags & __SO_ACCEPTCON)
		return -EINVAL;

	if (sock->state != SS_CONNECTED) 
		return -ENOTCONN;

	if (sk->state != CON_ESTABLISHED)  {
		if (sk->shutdown & SHUTDOWN_MASK) {
			//if (!(flags & MSG_NOSIGNAL))
			//	send_sig(SIGPIPE, current, 0);
			return sock_error(sk);
		}
		return -ENOTCONN;
	}


	skb = skb_recv_datagram(sk, (flags & ~MSG_DONTWAIT) | MSG_PEEK, flags & MSG_DONTWAIT, &err);
	if (skb == NULL)
		goto exit;

	DBPRT(DBRECV, "We received %d bytes\n", skb->len);
	DBPRTPKT(DBRECV, skb->data, skb->len);

	skb->h.raw = skb->data;

	if (size < skb->len) {
		copied = size;
		if (sock->type == SOCK_SEQPACKET)
			msg->msg_flags |= MSG_TRUNC;
	} else
		copied = skb->len;
	skb_copy_datagram_iovec(skb, 0, msg->msg_iov, copied);
	if (!(flags & MSG_PEEK)) {
		if (msg->msg_flags & MSG_TRUNC)
			skb->len = 0;
		else
			skb_pull(skb, copied);
	}
	if (!skb->len) {
		skb_unlink(skb);
		kfree_skb(skb);
	}
	kfree_skb(skb);		// remove MSG_PEEK link
	err = copied;
 exit:
	DBFEXIT;
	return err;
}


/*
    Protocol family operations
*/

struct proto_ops lpf_ops = {
	family:		PF_AFFIX,

	release:	lpf_release,
	bind:		lpf_bind,
	connect:	lpf_connect,
	socketpair:	sock_no_socketpair,
	accept:		lpf_accept,
	getname:	lpf_getname,
	poll:		datagram_poll,
	ioctl:		lpf_ioctl,
	listen:		lpf_listen,
	shutdown:	sock_no_shutdown,	/* it is used in ax25 */
	setsockopt:	lpf_setsockopt,
	getsockopt:	lpf_getsockopt,
	sendmsg:	lpf_sendmsg,
	recvmsg:	lpf_recvmsg,
	mmap:		sock_no_mmap
};

int __lpf_create(struct socket *sock, int protocol)
{
	struct sock		*sk;

	DBFENTER;

	sock->state = SS_UNCONNECTED;
	
	/* l2cap supports stream and seqpacket */
	if (sock->type != SOCK_STREAM && sock->type != SOCK_SEQPACKET)
		return -ESOCKTNOSUPPORT;

	sk = lpf_create(PF_AFFIX, GFP_KERNEL, 1);
	if (sk == NULL)
		return -ENOMEM;

	lpf_init_data(sock, sk);	/* generic initialization */
	sk->protocol = protocol;
	sock->ops = &lpf_ops;

	MOD_INC_USE_COUNT;

	DBFEXIT;
	return 0;
}


struct net_proto_family lpf_family_ops = {
	family:	PF_AFFIX,
	create:	__lpf_create
};

int lpf_proc_read(char *buf, char **start, off_t offset, int len)
{
	int count = 0;

	DBFENTER;

	count += sprintf(buf+count, "L2CAP sockets in use: %ld, list len: %d\n",
			sock_count, l2cap_socks.len);

	DBFEXIT;
	return count;
}


struct proc_dir_entry	*lpf_proc;

int init_lpf(void)
{
	int	err;
	DBFENTER;

	lpf_proc = create_proc_info_entry("l2cap_sock", 0, proc_affix, lpf_proc_read);
	if (lpf_proc == NULL) {
		BTERROR("Unable to register proc fs entry\n");
		return -EINVAL;
	}

	err = affix_sock_register(&lpf_family_ops, BTPROTO_L2CAP);
	if (err)
		goto err1;

	btl_head_init(&l2cap_socks);
	
	DBFEXIT;
	return 0;
 err1:
	remove_proc_entry("l2cap_sock", proc_affix);
	return err;
}

void exit_lpf(void)
{
	DBFENTER;

	affix_sock_unregister(BTPROTO_L2CAP);
	remove_proc_entry("l2cap_sock", proc_affix);

	DBFEXIT;
}

