/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: l2cap.c,v 1.27 2002/11/13 00:26:01 kds Exp $

   Link Layer Control and Adaptation Protocol

   Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
                Imre Deak <ext-imre.deak@nokia.com>
*/		


/* The following prevents "kernel_version" from being set in this file. */
#define __NO_VERSION__

#include <linux/config.h>

/* Module related headers, non-module drivers should not include */
#include <linux/module.h>
#include <linux/init.h>

/* Standard driver includes */
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/types.h>
#include <linux/timer.h>
#include <linux/errno.h>
#include <linux/spinlock.h>
#include <linux/proc_fs.h>
#include <linux/skbuff.h>
#include <asm/bitops.h>

/* Local Includes */
#define FILEBIT	DBL2CAP

#include <affix/bluetooth.h>
#include <affix/btdebug.h>
#include <affix/hci.h>
#include <affix/l2cap.h>


/* L2CAP layer attributes */
btlist_head_t	l2cap_protos;		/* upper protocol list */
btlist_head_t	l2cap_chs;		/* L2CAP channel list */
__u16		lcid = 0x40;
__u8		cmd_id = 0x01;


/************    CLIENT PROTOCOL MANAGEMENT ***************/


l2cap_proto *__l2cap_proto_lookup(__u16 psm)
{
	l2cap_proto	*proto;

	btl_for_each (proto, l2cap_protos)
		if (proto->psm == psm)
			break;
	return proto;
}

int l2ca_register_protocol(__u16 psm, l2cap_proto_ops *pops)
{
	int		err = -EINVAL;
	l2cap_proto	*proto;

	DBFENTER;

	btl_write_lock(&l2cap_protos);

	proto = __l2cap_proto_lookup(psm);
	if (proto == NULL) {
		proto = kmalloc(sizeof(*proto), GFP_ATOMIC);
		if (proto == NULL) {
			err = -ENOMEM;
			goto exit;
		}
		proto->psm = psm;
		proto->ops = pops;
		__btl_add_tail(&l2cap_protos, proto);
		err = 0;
	}

exit:
	btl_write_unlock(&l2cap_protos);

	DBFEXIT;
	return err;
}

void l2ca_unregister_protocol(__u16 psm)
{
	l2cap_proto	*proto;

	DBFENTER;
	btl_write_lock(&l2cap_protos);
	proto = __l2cap_proto_lookup(psm);
	if (proto) {
		__btl_unlink(&l2cap_protos, proto);
		kfree(proto);
	}

	btl_write_unlock(&l2cap_protos);
	DBFEXIT;
}

	
/*******************************   CHANNEL MANAGEMENT  ************************/

/*
   L2CAP timer
*/
void l2cap_timer(unsigned long data)
{
	l2cap_ch	*ch = (l2cap_ch*)data;

	DBFENTER;

	switch(SETSTATE(ch, DEAD)) {
		case CON_W4_CONRSP:
			l2ca_connect_cfm(ch, 0xEEEE, 0);
			break;
		case CON_CONFIG:
			if (test_and_clear_bit(L2CAP_CONFIG, &ch->flags))
				l2ca_config_cfm(ch, 0xEEEE);
			else
				l2ca_connect_cfm(ch, 0xEEEE, 0);
			break;
		case CON_W4_DISCRSP:
			//l2ca_disconnect_cfm(ch);
			break;
		default:
	}
	l2ca_put(ch);
}


/*
  probably the reason to use "con"
*/
l2cap_ch * __l2cap_lookup(__u16 cid)
{
	l2cap_ch	*ch;

	btl_for_each (ch, l2cap_chs)
		if (STATE(ch) != DEAD && ch->lcid == cid)
			return ch;

	return NULL;
}

l2cap_ch * l2cap_lookup(__u16 cid)
{
	l2cap_ch	*ch;

	btl_read_lock(&l2cap_chs);
	ch = __l2cap_lookup(cid);
	if (ch)
		l2ca_hold(ch);
	btl_read_unlock(&l2cap_chs);

	return ch;
}

l2cap_ch * __l2cap_lookup_id(int reqid)
{
	l2cap_ch	*ch;

	btl_for_each (ch, l2cap_chs)
		if (STATE(ch) != DEAD && ch->reqid == reqid)
			return ch;

	return NULL;
}

l2cap_ch * l2cap_lookup_id(int reqid)
{
	l2cap_ch	*ch;

	btl_read_lock(&l2cap_chs);
	ch = __l2cap_lookup_id(reqid);
	if (ch)
		l2ca_hold(ch);
	btl_read_unlock(&l2cap_chs);

	return ch;
}

static inline __u16 get_new_cid(void)
{
	__u16		cid, counter;
	l2cap_ch	*ch;

	DBFENTER;

	for (counter = 0x40; counter > 0 || (cid = CID_NULL); counter++) {
		cid = lcid++;
		if (lcid <= 0)
			lcid = 0x40;
		if ((ch = __l2cap_lookup(cid)) == NULL)
			break;
	}
	
	DBFEXIT;
	return cid;
}

static inline int bind_connection(l2cap_ch *ch, hci_con *con)
{
	if (ch->con)
		hcc_put(ch->con);	/* close old connection */

	hcc_hold(con);
	ch->con = con;
	ch->hdrlen = con->hdrlen + L2CAP_HDR_LEN;

	return 0;
}


/*
 * Create the channel object
 */

l2cap_ch *l2ca_create(__u16 psm, l2cap_proto_ops *pops)
{
	l2cap_ch	*ch;

	DBFENTER;

	ch = kmalloc(sizeof(l2cap_ch), GFP_ATOMIC);
	if (!ch)
		return NULL;
	memset(ch, 0, sizeof(l2cap_ch));
	
	/* Initialize some members */
	init_timer(&ch->timer);
	ch->timer.data = (unsigned long)ch;
	ch->timer.function = l2cap_timer;
	
	/* set protocol default values */
	def_cfg_opt(&ch->in_flow);
	def_cfg_opt(&ch->out_flow);

	ch->psm = psm;
	ch->ops = pops;
	atomic_set(&ch->refcnt, 1);

	ENTERSTATE(ch, CON_CLOSED);	/* initial state */

	btl_write_lock(&l2cap_chs);
	ch->lcid = get_new_cid();
	if (ch->lcid == CID_NULL) {
		kfree(ch);
		ch = NULL;
		goto exit;
	}
	__MOD_INC_USE_COUNT(pops->owner);
	__btl_add_tail(&l2cap_chs, ch);
exit:
	btl_write_unlock(&l2cap_chs);
	DBFEXIT;
	return ch;
}

l2cap_ch *l2cap_create(__u16 psm, int *err)
{
	l2cap_proto	*proto;
	l2cap_ch	*ch;

	btl_read_lock(&l2cap_protos);
	proto = __l2cap_proto_lookup(psm);
	if (proto) {
		*err = 0;
		ch = l2ca_create(proto->psm, proto->ops);
		if (ch != NULL)	/* No resources	*/
			*err = 0;
		else
			*err = -ENOMEM;
	} else {
		*err = -EPROTONOSUPPORT;
		ch = NULL;
	}
	btl_read_unlock(&l2cap_protos);

	return ch;
}

void __l2cap_destroy(l2cap_ch *ch)
{
	DBFENTER;

	if (atomic_read(&ch->refcnt) == 0) {
		__btl_unlink(&l2cap_chs, ch);
		if (ch->con)
			hcc_put(ch->con);
		if (ch->hci)
			hci_put(ch->hci);
		kfree(ch);
	}

	DBFEXIT;
}

void l2cap_destroy(l2cap_ch *ch)
{
	btl_write_lock(&l2cap_chs);
	__l2cap_destroy(ch);
	btl_write_unlock(&l2cap_chs);
}

static inline void __l2ca_put(l2cap_ch *ch)
{
	__MOD_DEC_USE_COUNT(ch->ops->owner);
	if (atomic_dec_and_test(&ch->refcnt))
		__l2cap_destroy(ch);
}

static inline int l2cap_unused(l2cap_ch *ch)
{
	if (atomic_read(&ch->refcnt) == 1 || ch->priv == NULL)
		return 1;
	return 0;
}

void l2ca_put(l2cap_ch *ch)
{
	DBFENTER;
	
	__MOD_DEC_USE_COUNT(ch->ops->owner);
	
	if (l2cap_unused(ch))
		l2ca_disconnect_req(ch);
	
	if (atomic_dec_and_test(&ch->refcnt))
		l2cap_destroy(ch);

	DBFEXIT;
}


/*
 * configuration stuff
 */
void def_cfg_opt(l2cap_cfg_opt *opt)
{
	opt->mtu = 672;
	opt->flush_to = 0xFFFF;
	opt->qos.flags = 0;
	opt->qos.service_type = QOS_BEST_EFFORT;
	opt->qos.token_rate = 0x00000000;
	opt->qos.token_size = 0x00000000;
	opt->qos.bandwidth = 0x00000000;
	opt->qos.latency = 0xFFFFFFFF;
	opt->qos.delay_variation = 0xFFFFFFFF;		
}

/* decode configuration options from packet to the object */
int decode_cfg_opt(struct sk_buff *skb, l2cap_cfg_opt *opt)
{
	l2cap_cfgopt_t 	*o;
	l2cap_qos_t	*qos;

	while(skb->len > 2) {
		o = (l2cap_cfgopt_t*)skb->data;
		
		switch (o->type) {
		case CFG_OPT_MTU:
			opt->mtu = __btoh16(__get_u16(o->data));
			set_bit(CFG_OPT_MTU, &opt->flags);
			skb_pull(skb, 4);
			break;
		case CFG_OPT_FLUSHTO:
			opt->flush_to = __btoh16(__get_u16(o->data));
			set_bit(CFG_OPT_FLUSHTO, &opt->flags);
			skb_pull(skb, 4);
			break;
		case CFG_OPT_QOS:
			qos = (l2cap_qos_t*)o->data;
			opt->qos.flags = qos->flags;
			opt->qos.service_type = qos->service_type;
			opt->qos.token_rate = __btoh32(qos->token_rate);
			opt->qos.token_size = __btoh32(qos->token_size);
			opt->qos.bandwidth = __btoh32(qos->bandwidth);
			opt->qos.latency = __btoh32(qos->latency);
			opt->qos.delay_variation = __btoh32(qos->delay_variation);		
			set_bit(CFG_OPT_QOS, &opt->flags);
			skb_pull(skb, 24);
			break;
		}

	}
	return 0;
}

/* encode configurations options from object to the packet */
int encode_cfg_opt(__u8 *options, l2cap_cfg_opt *opt)
{
	__u8		*optr = options;
	l2cap_cfgopt_t 	*o;
	l2cap_qos_t	*qos;

	/* Construct options buffer */
	if (test_bit(CFG_OPT_MTU, &opt->flags)) {
		o = (l2cap_cfgopt_t*)optr;
		o->type = CFG_OPT_MTU;
		o->length = 2;
		__put_u16(o->data, __htob16(opt->mtu));
		optr += 4;
	}
	if (test_bit(CFG_OPT_FLUSHTO, &opt->flags)) {
		o = (l2cap_cfgopt_t*)optr;
		o->type = CFG_OPT_FLUSHTO;
		o->length = 2;
		__put_u16(o->data, __htob16(opt->flush_to));
		optr += 4;
	}
	if (test_bit(CFG_OPT_QOS, &opt->flags)) {
		o = (l2cap_cfgopt_t*)optr;
		o->type = CFG_OPT_QOS;
		o->length = 22;
		qos = (l2cap_qos_t*)o->data;
		qos->flags = opt->qos.flags;
		qos->service_type = opt->qos.service_type;
		qos->token_rate = __htob32(opt->qos.token_rate);
		qos->token_size = __htob32(opt->qos.token_size);
		qos->bandwidth = __htob32(opt->qos.bandwidth);
		qos->latency = __htob32(opt->qos.latency);
		qos->delay_variation = __htob32(opt->qos.delay_variation);
		optr += 24;
	}
	return optr - options;
}

/****************************  SIGNALING and DATA SENDING  ***********************/

int l2cap_send_data(l2cap_ch *ch, struct sk_buff *skb)
{
	__u16		len = skb->len;
	l2cap_hdr_t	*hdr;

	DBFENTER;

	hdr = (l2cap_hdr_t*)skb_push(skb, L2CAP_HDR_LEN);
	hdr->length = __htob16(len);
	hdr->cid = __htob16(ch->rcid);

	DBPRT(DBXMIT, "Send L2CAP Packet, len = %d\n", skb->len);

	hci_send(ch->con, skb);

	DBFEXIT;
	return 0;
}

int l2cap_send_signal(hci_con *con, struct sk_buff *skb)
{
	__u16		len = skb->len;
	l2cap_hdr_t	*hdr;

	DBFENTER;

	hdr = (l2cap_hdr_t*)skb_push(skb, L2CAP_HDR_LEN);
	hdr->length = __htob16(len);
	hdr->cid = __htob16(CID_SIGNAL);

	DBPRT(DBXMIT, "Send L2CAP Packet, len = %d\n", skb->len);
	DBPRTPKT(DBXMIT, skb->data, skb->len);

	hci_send(con, skb);

	DBFEXIT;
	return 0;
}

/* Configuration helpers */
int l2p_config_req(l2cap_ch *ch)
{
	int	result = 0;
	__u8	options[32], olen, more = 0;
	__u16	flags;

	DBFENTER;

	// check if we already sent request (out_state) ???
	if (ch->out_state == CHN_CFG_START) {
		flags = more?CFG_REQ_MORE:0;
		olen = encode_cfg_opt(options, &ch->out_flow);
		result = l2cap_config_req(ch->con, get_id(ch), ch->rcid, flags, olen, options);
		/* start timer */
		l2cap_start_timer(ch, RTX_TIMEOUT);
		/* change state */
		ch->out_state = CHN_CFG_REQSENT;	/*???*/
	}


	DBFEXIT;
	return result;
}

int l2p_config_rsp(l2cap_ch *ch, __u16 result)
{
	int	err = 0;
	__u8	options[32], olen, more = 0;
	__u16	flags;

	DBFENTER;

	flags = more?CFG_REQ_MORE:0;
	olen = encode_cfg_opt(options, &ch->in_flow);;
	err = l2cap_config_rsp(ch->con, ch->rspid, ch->rcid, flags, result, olen, options);
	if (result == CFG_RSP_SUCCESS) {
		ch->in_state = CHN_CFG_END;
		if (ch->out_state == CHN_CFG_END) {
			ENTERSTATE(ch, CON_OPEN);
			if (test_and_clear_bit(L2CAP_CONFIG, &ch->flags))
				l2ca_config_cfm(ch, result);
			else
				l2ca_connect_cfm(ch, result, 0);
		}
	} else
		ch->in_state = CHN_CFG_START;


	DBFEXIT;
	return err;
}



/* handler functions */

int sig_reject(hci_con *con, struct sk_buff *skb)
{
#ifdef CONFIG_AFFIX_DEBUG
	l2cap_cmd_t	*cmd = (l2cap_cmd_t*)skb->data;
	__u8		id = cmd->id;

	DBFENTER;

	/* Do nothing */
	DBPRT(DBCTRL, "We have rejected command, id: %d\n", id);
#endif
	
	kfree_skb(skb);
	return 0;
}

int sig_connect_req(hci_con *con, struct sk_buff *skb)
{
	__u16		rcid;
	__u16		psm;
	l2cap_conreq_t	*req = (l2cap_conreq_t*)skb->data;
	l2cap_ch	*ch;
	int		err;
	
	DBFENTER;

	rcid = __btoh16(req->scid);
	psm = __btoh16(req->psm);
	ch = l2cap_create(psm, &err);
	if (ch == NULL) {
		if (err == -ENOMEM)
			l2cap_connect_rsp(con, req->id, rcid, CID_NULL, CON_RSP_RESOURCE, CON_RSP_NOINFO);
		else
			/* Send response with unknown PSM options */
			l2cap_connect_rsp(con, req->id, rcid, CID_NULL, CON_RSP_PSM, CON_RSP_NOINFO);
		goto exit;
	}
	
	bind_connection(ch, con);
	set_bit(L2CAP_SERVER, &ch->flags);
	ch->rspid = req->id;
	ch->rcid = rcid;
	ch->bda = con->bda;
	hci_hold(con->hci);	// set reference
	ch->hci = con->hci;

	ENTERSTATE(ch, CON_W4_LCONRSP);
	l2ca_connect_ind(ch);
	l2ca_put(ch);

 exit:
	kfree_skb(skb);
	DBFEXIT;
	return 0;
}

int sig_connect_rsp(hci_con *con, struct sk_buff *skb)
{
	__u16		rcid, lcid;
	l2cap_ch	*ch;
	l2cap_conrsp_t	*rsp = (l2cap_conrsp_t*)skb->data;
	__u16		result, status;
	
	DBFENTER;

	rcid = __btoh16(rsp->dcid);
	lcid = __btoh16(rsp->scid);
	result = __btoh16(rsp->result);
	status = __btoh16(rsp->status);
	
	ch = l2cap_lookup(lcid);
	if (ch == NULL)
		goto exit;

	l2cap_stop_timer(ch);	// stop timer

	if (STATE(ch) == CON_W4_CONRSP) {
		switch (result) {
		case CON_RSP_SUCCESS:
			ch->rcid = rcid;
			l2cap_config_start(ch);
			l2p_config_req(ch);
			break;

		case CON_RSP_PENDING:
			l2cap_start_timer(ch, ERTX_TIMEOUT);
			break;

		default:
			/* CON_RSP_NEG */
			ENTERSTATE(ch, DEAD);
			l2ca_connect_cfm(ch, result, status);
		}
	}
	l2ca_put(ch);
 exit:
	kfree_skb(skb);
	DBFEXIT;
	return 0;
}

int sig_config_req(hci_con *con, struct sk_buff *skb)
{
	l2cap_cfgreq_t	*req = (l2cap_cfgreq_t*)skb->data;
	__u16		lcid = __btoh16(req->dcid);
	l2cap_ch	*ch;

	DBFENTER;

	ch = l2cap_lookup(lcid);
	if (ch == NULL) {
		l2cap_command_rej(con, req->id, REJ_INVCID, 4, (__u8*)&lcid);
		goto exit;
	}

	switch (STATE(ch)) {
	case CON_CLOSED:
		break;
	case CON_OPEN:
		/* Here I have to disable data transmission as well ??? */
		set_bit(L2CAP_CONFIG, &ch->flags);
		l2cap_config_start(ch);
	case CON_CONFIG:
		ch->rspid = req->id;
		if (ch->in_state == CHN_CFG_START) {
			/* stop dead config timer */
			if (ch->out_state == CHN_CFG_END)
				l2cap_stop_timer(ch);
			ch->in_state = CHN_CFG_REQRECV;
			skb_pull(skb, 8);
			decode_cfg_opt(skb, &ch->in_flow);
			// we have to check if parameters are OK
			if (test_bit(L2CAP_CONFIG, &ch->flags) && ch->out_state == CHN_CFG_START)
				l2ca_config_ind(ch);
			else
				l2p_config_rsp(ch, 0);	//XXX: just responding
		}
		break;
	default:
		DBPRT(DBCTRL, "State is not processed\n");
	}

	l2ca_put(ch);
 exit:
	kfree_skb(skb);
	DBFEXIT;
	return 0;
}

int sig_config_rsp(hci_con *con, struct sk_buff *skb)
{
	__u16		lcid;
	l2cap_ch	*ch;
	l2cap_cfgrsp_t	*rsp = (l2cap_cfgrsp_t*)skb->data;
	__u16		result = __btoh16(rsp->result);

	DBFENTER;

	lcid = __btoh16(rsp->scid);

	ch = l2cap_lookup(lcid);
	if (ch == NULL) {
		l2cap_command_rej(con, rsp->id, REJ_INVCID, 4, (__u8*)&lcid);
		goto exit;
	}

	l2cap_stop_timer(ch);	// STOP TIMER

	if (STATE(ch) == CON_CONFIG) {
		if (ch->out_state == CHN_CFG_REQSENT) {
			skb_pull(skb, 10);
			decode_cfg_opt(skb, &ch->out_flow);
			if (result == CFG_RSP_SUCCESS) {
				ch->out_state = CHN_CFG_END;
				if (ch->in_state == CHN_CFG_END) {
					ENTERSTATE(ch, CON_OPEN);
					if (test_and_clear_bit(L2CAP_CONFIG, &ch->flags))
						l2ca_config_cfm(ch, result);
					else
						l2ca_connect_cfm(ch, result, 0);
				} else {
					/* start dead config timer
					 * e.g. other side can hang
					 */
					if (ch->in_state == CHN_CFG_START)
						l2cap_start_timer(ch, RTX_TIMEOUT);
				}
			} else
				ch->out_state = CHN_CFG_START;
		}
	}

	l2ca_put(ch);
 exit:
	kfree_skb(skb);
	DBFEXIT;
	return 0;
}

int sig_disconnect_req(hci_con *con, struct sk_buff *skb)
{
	__u16		rcid, lcid;
	l2cap_discreq_t	*req = (l2cap_discreq_t*)skb->data;
	l2cap_ch	*ch;
			
	DBFENTER;

	rcid = __btoh16(req->scid);
	lcid = __btoh16(req->dcid);

	ch = l2cap_lookup(lcid);
	if (ch == NULL || ch->rcid != rcid) {
		l2cap_command_rej(con, req->id, REJ_INVCID, 4, (__u8*)&lcid);
		if (ch)
			l2ca_put(ch);
		goto exit;
	}

	/*
	  Check first if we sent disconnect_req already
	*/	  
	if (STATE(ch) == CON_CLOSED || STATE(ch) == CON_W4_DISCRSP) {
		l2cap_disconnect_rsp(ch->con, req->id, rcid, lcid);
	} else {
		ch->rspid = req->id;
		ENTERSTATE(ch, DEAD);
		l2ca_disconnect_ind(ch);
		l2cap_disconnect_rsp(ch->con, ch->rspid, ch->rcid, ch->lcid);
	}
	l2ca_put(ch);
 exit:
	kfree_skb(skb);
	DBFEXIT;
	return 0;
}

int sig_disconnect_rsp(hci_con *con, struct sk_buff *skb)
{
	__u16		lcid;
	l2cap_discrsp_t	*rsp = (l2cap_discrsp_t*)skb->data;
	l2cap_ch	*ch;
			
	DBFENTER;

	lcid = __btoh16(rsp->scid);

	ch = l2cap_lookup(lcid);
	if (ch == NULL) {
		l2cap_command_rej(con, rsp->id, REJ_INVCID, 4, (__u8*)&lcid);
		goto exit;
	}

	l2cap_stop_timer(ch);

	if (STATE(ch) == CON_W4_DISCRSP) {
		ENTERSTATE(ch, DEAD);
	}
	l2ca_put(ch);
 exit:
	kfree_skb(skb);
	DBFEXIT;
	return 0;
}

int sig_echo_req(hci_con *con, struct sk_buff *skb)
{
	l2cap_cmd_t	*cmd;
	struct sk_buff	*rsp;

	DBFENTER;
	
	rsp = skb_realloc_headroom(skb, con->hdrlen + L2CAP_HDR_LEN);
	if (rsp) {
		/* send the same data back */
		cmd = (void*)rsp->data;
		cmd->code = SIG_ECHORSP;
		l2cap_send_signal(con, rsp);
	}

	kfree_skb(skb);
	DBFEXIT;
	return 0;
}

int sig_echo_rsp(hci_con *con, struct sk_buff *skb)
{
	l2cap_cmd_t	*req = (void*)skb->data;
	//int		len = __btoh16(req->length);
	l2cap_ch	*ch;

	DBFENTER;

	//kfree_skb(skb);
	//return 0;

	ch = l2cap_lookup_id(req->id);
	if (ch == NULL) {
		kfree_skb(skb);
		return 0;
	}
	
	if (STATE(ch) == CON_OPEN) {
		skb_pull(skb, 4);
		l2ca_data_ind(ch, skb);
	} else {
		DBPRT(DBRECV, "Channel closed\n");
		kfree_skb(skb);
	}
	l2ca_put(ch);

	DBFEXIT;
	return 0;
}


int sig_info_req(hci_con *con, struct sk_buff *skb)
{
	l2cap_inforeq_t	*req = (void*)skb->data;
	__u16		mtu = __htob16(672);

	DBFENTER;

	l2cap_info_rsp(con, req->id, INFO_REQ_MTU, 2, (__u8*)&mtu);

	kfree_skb(skb);
	DBFEXIT;
	return 0;
}

void signal_handler(hci_con *con, struct sk_buff *skb)
{
	struct sk_buff	*cskb;
	__u8		code;
	__u16		cmdlen;
	l2cap_cmd_t	*cmd = (l2cap_cmd_t*)skb->data;

	DBFENTER;

	if (skb->len < 4) /* less then hdrlen */
		goto exit;
	cmdlen = __btoh16(cmd->length) + 4;	// +4 .. because of header of the command


	while (skb->len >= cmdlen) {	/* Process several commands	*/
		code = cmd->code;

		cskb = skb_clone(skb, GFP_ATOMIC);
		skb_trim(cskb, cmdlen);

		switch (code) {
		case SIG_REJECT:
			sig_reject(con, cskb);
			break;

		case SIG_CONREQ:
			DBPRT(DBCTRL, "SIGNAL: CONNECT_REQ\n");
			sig_connect_req(con, cskb);
			break;

		case SIG_CONRSP:
			DBPRT(DBCTRL, "SIGNAL: CONNECT_RSP\n");
			sig_connect_rsp(con, cskb);
			break;

		case SIG_CFGREQ:
			DBPRT(DBCTRL, "SIGNAL: CONFIG_REQ\n");
			sig_config_req(con, cskb);
			break;

		case SIG_CFGRSP:
			DBPRT(DBCTRL, "SIGNAL: CONFIG_RSP\n");
			sig_config_rsp(con, cskb);
			break;

		case SIG_DISCREQ:
			DBPRT(DBCTRL, "SIGNAL: DISCONNECT_REQ\n");
			sig_disconnect_req(con, cskb);
			break;

		case SIG_DISCRSP:
			DBPRT(DBCTRL, "SIGNAL: DISCONNECT_RSP\n");
			sig_disconnect_rsp(con, cskb);
			break;

		case SIG_ECHOREQ:
			DBPRT(DBCTRL, "SIGNAL: ECHO_REQ\n");
			sig_echo_req(con, cskb);
			break;

		case SIG_ECHORSP:
			DBPRT(DBCTRL, "SIGNAL: ECHO_RSP\n");
			sig_echo_rsp(con, cskb);
			break;

		case SIG_INFOREQ:
			DBPRT(DBCTRL, "SIGNAL: INFO_REQ\n");
			sig_info_req(con, cskb);
			break;

		case SIG_INFORSP:
			DBPRT(DBCTRL, "SIGNAL: INFO_RSP - unprocessed\n");
			kfree_skb(cskb);
			break;

		case SIG_SINGLEPING:
			kfree_skb(cskb);
			break;

		default:
			DBPRT(DBCTRL, "Unknown command: %#02x\n", code);
			kfree_skb(cskb);
			l2cap_command_rej(con, code, REJ_NOTSUP, 0, NULL);
		}
		skb_pull(skb, cmdlen);
		if (skb->len < 4)
			break;
		cmd = (l2cap_cmd_t*)skb->data;
		cmdlen = __btoh16(cmd->length) + 4;
	}

 exit:
	kfree_skb(skb);
	DBFEXIT;
}

void group_handler(hci_con *con, struct sk_buff *skb)
{
	DBFENTER;

	DBPRT(DBCTRL, "Group addressing is not supported yet\n");

	kfree_skb(skb);

	DBFEXIT;
}

void channel_handler(hci_con *con, __u16 cid, struct sk_buff *skb)
{
	l2cap_ch	*ch;

	DBFENTER;

	ch = l2cap_lookup(cid);
	if (ch == NULL) {
		kfree_skb(skb);
		return;
	}
	
	if (STATE(ch) == CON_OPEN) {
		l2ca_data_ind(ch, skb);
	} else {
		DBPRT(DBRECV, "Channel closed\n");
		kfree_skb(skb);
	}
	l2ca_put(ch);

	DBFEXIT;
}


/********************   Signalling commands   ***************************/

int l2cap_command_rej(hci_con *con, __u8 id, __u16 reason, __u16 length, __u8 *data)
{
	struct sk_buff		*skb;
	l2cap_cmdrej_t		*cmd;

	DBFENTER;

	skb = alloc_skb(con->hdrlen+L2CAP_HDR_LEN+6+length, GFP_ATOMIC);
	if (skb == NULL) return -1;
	skb_reserve(skb, con->hdrlen+L2CAP_HDR_LEN);
	skb_put(skb, 6);

	cmd = (l2cap_cmdrej_t*)skb->data;
	cmd->code = SIG_REJECT;
	cmd->id = id;
	cmd->length = __htob16(2 + length);
	cmd->reason = __htob16(reason);
	if (length)
		memcpy(skb_put(skb, length), data, length);

	l2cap_send_signal(con, skb);

	DBFEXIT;
	return 0;
}

int l2cap_connect_req(hci_con *con, __u8 id, __u16 psm, __u16 lcid)
{
	struct sk_buff		*skb;
	l2cap_conreq_t		*cmd;


	DBFENTER;

	skb = alloc_skb(con->hdrlen+L2CAP_HDR_LEN+8, GFP_ATOMIC);
	if (skb == NULL) return -1;
	skb_reserve(skb, con->hdrlen+L2CAP_HDR_LEN);
	skb_put(skb, 8);

	cmd = (l2cap_conreq_t*)skb->data;
	cmd->code = SIG_CONREQ;
	cmd->id = id;
	cmd->length = __htob16(4);
	cmd->psm = __htob16(psm);
	cmd->scid = __htob16(lcid);

	l2cap_send_signal(con, skb);

	DBFEXIT;
	return 0;
}

int l2cap_connect_rsp(hci_con *con, __u8 id, __u16 rcid, __u16 lcid, __u16 result, __u16 status)
{
	struct sk_buff		*skb;
	l2cap_conrsp_t		*cmd;

	DBFENTER;

	skb = alloc_skb(con->hdrlen+L2CAP_HDR_LEN+12, GFP_ATOMIC);
	if (skb == NULL) return -1;
	skb_reserve(skb, con->hdrlen+L2CAP_HDR_LEN);
	skb_put(skb, 12);
	
	cmd = (l2cap_conrsp_t*)skb->data;
	cmd->code = SIG_CONRSP;
	cmd->id = id;
	cmd->length = __htob16(8);
	cmd->scid = __htob16(rcid);
	cmd->dcid = __htob16(lcid);
	cmd->result = __htob16(result);
	cmd->status = __htob16(status);

	l2cap_send_signal(con, skb);

	DBFEXIT;
	return 0;
}

int l2cap_config_req(hci_con *con, __u8 id, __u16 rcid, __u16 flags, __u8 olen, __u8 *options)
{
	struct sk_buff		*skb;
	l2cap_cfgreq_t		*cmd;

	DBFENTER;

	skb = alloc_skb(con->hdrlen+L2CAP_HDR_LEN+8+olen, GFP_ATOMIC);
	if (skb == NULL) return -1;
	skb_reserve(skb, con->hdrlen+L2CAP_HDR_LEN);
	skb_put(skb, 8);
	
	cmd = (l2cap_cfgreq_t*)skb->data;
	cmd->code = SIG_CFGREQ;
	cmd->id = id;
	cmd->length = __htob16(4+olen);
	cmd->dcid = __htob16(rcid);
	cmd->flags = __htob16(flags);
	memcpy(skb_put(skb, olen), options, olen);

	l2cap_send_signal(con, skb);

	DBFEXIT;
	return 0;
}

int l2cap_config_rsp(hci_con *con, __u8 id, __u16 rcid, __u16 flags, __u16 result, __u8 olen, __u8 *options)
{
	struct sk_buff		*skb;
	l2cap_cfgrsp_t		*cmd;

	DBFENTER;

	skb = alloc_skb(con->hdrlen+L2CAP_HDR_LEN+10+olen, GFP_ATOMIC);
	if (skb == NULL) return -1;
	skb_reserve(skb, con->hdrlen+L2CAP_HDR_LEN);
	skb_put(skb, 10);

	cmd = (l2cap_cfgrsp_t*)skb->data;
	cmd->code = SIG_CFGRSP;
	cmd->id = id;
	cmd->length = __htob16(6+olen);
	cmd->scid = __htob16(rcid);
	cmd->flags = __htob16(flags);
	cmd->result = __htob16(result);
	memcpy(skb_put(skb, olen), options, olen);

	l2cap_send_signal(con, skb);

	DBFEXIT;
	return 0;
}

int l2cap_disconnect_req(hci_con *con, __u8 id, __u16 rcid, __u16 lcid)
{
	struct sk_buff		*skb;
	l2cap_discreq_t		*cmd;

	DBFENTER;

	skb = alloc_skb(con->hdrlen+L2CAP_HDR_LEN+8, GFP_ATOMIC);
	if (skb == NULL) return -1;
	skb_reserve(skb, con->hdrlen+L2CAP_HDR_LEN);
	skb_put(skb, 8);

	cmd = (l2cap_discreq_t*)skb->data;
	cmd->code = SIG_DISCREQ;
	cmd->id = id;
	cmd->length = __htob16(4);
	cmd->dcid = __htob16(rcid);
	cmd->scid = __htob16(lcid);

	l2cap_send_signal(con, skb);

	DBFEXIT;
	return 0;
}

int l2cap_disconnect_rsp(hci_con *con, __u8 id, __u16 rcid, __u16 lcid)
{
	struct sk_buff		*skb;
	l2cap_discrsp_t		*cmd;

	DBFENTER;

	skb = alloc_skb(con->hdrlen+L2CAP_HDR_LEN+8, GFP_ATOMIC);
	if (skb == NULL) return -1;
	skb_reserve(skb, con->hdrlen+L2CAP_HDR_LEN);
	skb_put(skb, 8);

	cmd = (l2cap_discrsp_t*)skb->data;
	cmd->code = SIG_DISCRSP;
	cmd->id = id;
	cmd->length = __htob16(4);
	cmd->dcid = __htob16(lcid);
	cmd->scid = __htob16(rcid);

	l2cap_send_signal(con, skb);

	DBFEXIT;
	return 0;
}

int l2cap_echo_req(hci_con *con, __u8 id, struct sk_buff *skb, __u8 code)
{
	l2cap_cmd_t	*cmd;

	DBFENTER;

	cmd = (l2cap_cmd_t*)skb_push(skb, 4);
	cmd->code = code;
	cmd->id = id;
	cmd->length = __htob16(skb->len - 4);

	l2cap_send_signal(con, skb);

	DBFEXIT;
	return 0;
}

int l2cap_info_req(hci_con *con, __u8 id, __u16 infotype)
{
	struct sk_buff		*skb;
	l2cap_inforeq_t		*cmd;

	DBFENTER;

	skb = alloc_skb(con->hdrlen+L2CAP_HDR_LEN+6, GFP_ATOMIC);
	if (skb == NULL) return -1;
	skb_reserve(skb, con->hdrlen+L2CAP_HDR_LEN);
	skb_put(skb, 6);

	cmd = (l2cap_inforeq_t*)skb->data;
	cmd->code = SIG_INFOREQ;
	cmd->id = id;
	cmd->length = __htob16(2);
	cmd->infotype = __htob16(infotype);

	l2cap_send_signal(con, skb);

	DBFEXIT;
	return 0;
}

int l2cap_info_rsp(hci_con *con, __u8 id, __u16 infotype, __u16 length, __u8 *data)
{
	struct sk_buff		*skb;
	l2cap_inforsp_t		*cmd;

	DBFENTER;

	skb = alloc_skb(con->hdrlen+L2CAP_HDR_LEN+8+length, GFP_ATOMIC);
	if (skb == NULL) return -1;
	skb_reserve(skb, con->hdrlen+L2CAP_HDR_LEN);
	skb_put(skb, 8);

	cmd = (l2cap_inforsp_t*)skb->data;
	cmd->code = SIG_INFORSP;
	cmd->id = id;
	cmd->length = __htob16(4+length);
	cmd->infotype = __htob16(infotype);
	memcpy(skb_put(skb, length), data, length);

	l2cap_send_signal(con, skb);

	DBFEXIT;
	return 0;
}


/****************************   Upper to L2CAP services   ************************/


int l2ca_connect_req(l2cap_ch *ch, BD_ADDR *bda)
{
	int		err = 0;
	hci_con		*con;
	
	DBFENTER;

	/* set some attributes */
	ch->bda = *bda;
	ch->rcid = CID_NULL;

	/* Create Baseband connection */
	err = lp_connect_req(ch->hci, bda, &con);
	if (err) {
		DBPRT(DBXMIT, "Unable to create HCI connection\n");
		return err;
	}

	bind_connection(ch, con);
	hcc_put(con);
	
	ENTERSTATE(ch, CON_CLOSED);

	/* check hci con */
	if (!hcc_linkup(ch->con))
		return 0;

	ch->hdrlen = con->hdrlen + L2CAP_HDR_LEN;

	if (STATE(ch) == CON_CLOSED) {
		if (ch->psm) {
			/* connection oriented */
			ENTERSTATE(ch, CON_W4_CONRSP);
			l2cap_connect_req(ch->con, get_id(ch), ch->psm, ch->lcid);
			l2cap_start_timer(ch, RTX_TIMEOUT);
		} else {
			/* connectionless */
			ENTERSTATE(ch, CON_OPEN);
			l2ca_connect_cfm(ch, 0, 0);
		}
	}

	DBFEXIT;
	return 0;
}

int l2ca_connect_rsp(l2cap_ch *ch, __u16 result, __u16 status)
{
	DBFENTER;

	if (STATE(ch) == CON_W4_LCONRSP) {
		l2cap_connect_rsp(ch->con, ch->rspid, ch->rcid, ch->lcid, result, status);
		if (result == CON_RSP_SUCCESS) {
			l2cap_config_start(ch);
			l2p_config_req(ch);
		} else {
			if (result >= CON_RSP_PSM) {
				ENTERSTATE(ch, DEAD);
			}
		}
		return 0;
	}

	DBFEXIT;
	return -1;
}

/*
  before issueing configuration request
  ones should set configuration options on the object
*/
int l2ca_config_req(l2cap_ch *ch)
{
	int	result = 0;

	DBFENTER;

	switch (STATE(ch)) {
	case CON_CLOSED:
		l2ca_config_cfm(ch, 0xFFFF);	// ConfigCfmNeg
		break;
	case CON_OPEN:
		/* Here I have to disable data transmission as well ??? */
		set_bit(L2CAP_CONFIG, &ch->flags);
		l2cap_config_start(ch);
	case CON_CONFIG:
		result = l2p_config_req(ch);
		break;
	default:
	}

	DBFEXIT;
	return result;
}

int l2ca_config_rsp(l2cap_ch *ch, __u16 result)
{
	int	err = 0;

	DBFENTER;

	if( STATE(ch) == CON_CONFIG) {
		if (ch->in_state == CHN_CFG_REQRECV) {
			err = l2p_config_rsp(ch, result);
			if (result == CFG_RSP_SUCCESS)
				l2p_config_req(ch);
		}
	}

	DBFEXIT;
	return err;
}

int l2ca_disconnect_req(l2cap_ch *ch)
{
	int	err = 0;

	DBFENTER;
	if (STATE(ch) == CON_OPEN || STATE(ch) == CON_CONFIG) {
		if (ch->psm) {
			ENTERSTATE(ch, CON_W4_DISCRSP);
			l2cap_disconnect_req(ch->con, get_id(ch), ch->rcid, ch->lcid);
			/* start timer */
			l2cap_start_timer(ch, RTX_TIMEOUT_DISC);
		} else {
			/* connectionless */
			ENTERSTATE(ch, DEAD);
		}
	} else if (STATE(ch) == CON_W4_CONRSP)
		l2cap_start_timer(ch, RTX_TIMEOUT_DISC);
		
	DBFEXIT;
	return err;
}

int l2ca_send_data(l2cap_ch *ch, struct sk_buff *skb)
{
	int	err = 0;

	DBFENTER;

	if (STATE(ch) != CON_OPEN) {
		DBPRT(DBXMITE, "We are not in OPEN state, state: %d\n", STATE(ch));
		err = -ENOTCONN;
		goto exit;
	}

	if (skb_headroom(skb) < ch->hdrlen) {
		DBPRT(DBXMITE, "Headroom is small: %d\n", skb_headroom(skb));
		err = -ENOTCONN;
		goto exit;
	}
	DBPRT(DBXMITE, "Headroom is: %d\n", skb_headroom(skb));
	err = l2cap_send_data(ch, skb);

 exit:
	DBFEXIT;
	return err;
}

int l2ca_ping(l2cap_ch *ch, struct sk_buff *skb)
{
	int	err = 0;

	DBFENTER;

	if (STATE(ch) != CON_OPEN) {
		DBPRT(DBXMITE, "We are not in OPEN state, state: %d\n", STATE(ch));
		err = -ENOTCONN;
		goto exit;
	}

	err = l2cap_echo_req(ch->con, get_id(ch), skb, SIG_ECHOREQ);

 exit:
	DBFEXIT;
	return err;
}

int l2ca_singleping(l2cap_ch *ch, struct sk_buff *skb)
{
	int	err = 0;

	DBFENTER;

	if (STATE(ch) != CON_OPEN) {
		DBPRT(DBXMITE, "We are not in OPEN state, state: %d\n", STATE(ch));
		err = -ENOTCONN;
		goto exit;
	}
	err = l2cap_echo_req(ch->con, get_id(ch), skb, SIG_SINGLEPING);
 exit:
	DBFEXIT;
	return err;
}



int l2ca_connect_ind(l2cap_ch *ch)
{
	return ch->ops->connect_ind(ch);
}

int l2ca_connect_cfm(l2cap_ch *ch, __u16 result, __u16 status)
{
	if (ch->priv)
		return ch->ops->connect_cfm(ch, result, status);
	return 0;
}

int l2ca_config_ind(l2cap_ch *ch)
{
	if (ch->priv)
		return ch->ops->config_ind(ch);
	return 0;
}

int l2ca_config_cfm(l2cap_ch *ch, __u16 result)
{
	if (ch->priv)
		return ch->ops->config_cfm(ch, result);
	return 0;
}

int l2ca_disconnect_ind(l2cap_ch *ch)
{
	if (ch->priv)
		return ch->ops->disconnect_ind(ch);
	return 0;
}

int l2ca_data_ind(l2cap_ch *ch, struct sk_buff *skb)
{
	if (ch->priv)
		return ch->ops->data_ind(ch, skb);
	return 0;
}


/******************    Lower layer helper functions    *********************/

/*
 * Called from HCI layer when new event arrive
 */
struct sk_buff * lp_reassemble_packet(hci_con *con, __u8 first, struct sk_buff *skb)
{
	__u16		length;
	struct sk_buff	*s;

	DBFENTER;

	if (first) {
		if (con->rx_skb)
			BTDEBUG("con->rx_skb is not NULL, len: %d\n", con->rx_skb->len);

		if (skb->len < L2CAP_HDR_LEN) {
			BTDEBUG("L2CAP packet too small, drop it...\n");
			goto exit;
		}
		length = __btoh16(__get_u16(skb->data));
		DBPRT(DBRECV, "Real length: %d\n", length + L2CAP_HDR_LEN);
		if (skb->len >= length + L2CAP_HDR_LEN)
			return skb;
		
		s = alloc_skb(length + L2CAP_HDR_LEN, GFP_ATOMIC);
		if (s == NULL)
			goto exit;
		
		memcpy(skb_put(s, skb->len), skb->data, skb->len);
		if (con->rx_skb)
			kfree_skb(con->rx_skb);
		con->rx_skb = s;		

	} else {
		s = con->rx_skb;
		if (s == NULL) goto exit;

		length = __btoh16(__get_u16(s->data));

		if (skb_tailroom(s) < skb->len) {
			DBPRT(DBRECV, "Received packet too big\n");
			con->rx_skb = NULL;
			kfree_skb(s);
			goto exit;
		}

		memcpy(skb_put(s, skb->len), skb->data, skb->len);
		DBPRT(DBRECV, "accumulated, len: %d\n", s->len);
		if (s->len >= length + L2CAP_HDR_LEN) {
			con->rx_skb = NULL;
			kfree_skb(skb);
			DBFEXIT;
			return s;
		}
	}
	
 exit:
	kfree_skb(skb);
	DBFEXIT;
	return NULL;
}

int lp_receive_data(hci_con *con, __u8 first, __u8 bc, struct sk_buff *skb)
{
	__u16			cid;
	__u16			len;
	l2cap_grp_hdr_t		*hdr;

	DBFENTER;

	skb = lp_reassemble_packet(con, first, skb);
	if (skb == NULL)
		return 0;	/* accumulated */

	DBPRT(DBRECV, "L2CAP packet received\n");
	DBPRTPKT(DBRECV, skb->data, skb->len);

	hdr = (l2cap_grp_hdr_t*)skb->data;
	len = __btoh16(hdr->length);
	cid = __btoh16(hdr->cid);

	skb_pull(skb, L2CAP_HDR_LEN);

	if (skb->len < len) { /* wrong packet */
		BTDEBUG("L2CAP packet too small, drop it...\n");
		kfree_skb(skb);
		return 0;
	}

	switch (cid) {
		case CID_SIGNAL:
			signal_handler(con, skb);
			break;

		case CID_GROUP:
			group_handler(con, skb);
			break;

		default:	/* Dynamic channels */
			channel_handler(con, cid, skb);
	}
	DBFEXIT;
	return 0;
}

/*
 * from HCI
 * send l2cap connection request
 */
int lp_connect_cfm(hci_con *con, __u8 status)
{
	l2cap_ch	*ch, *next;
	int		err = -1, rsp;
	DBFENTER;

	btl_write_lock(&l2cap_chs);
	btl_for_each_next (ch, l2cap_chs, next) {
		l2ca_hold(ch);
		if (ch->con == con) {
			if (status == 0) {
				ch->hdrlen = con->hdrlen + L2CAP_HDR_LEN;
				if (STATE(ch) == CON_CLOSED) {
					if (ch->psm) {
						/* connection oriented */
						ENTERSTATE(ch, CON_W4_CONRSP);
						l2cap_connect_req(ch->con, get_id(ch), ch->psm, ch->lcid);
						l2cap_start_timer(ch, RTX_TIMEOUT);
					} else {
						/* connectionless */
						ENTERSTATE(ch, CON_OPEN);
						l2ca_connect_cfm(ch, 0, 0);
					}
				}
			} else {
				ENTERSTATE(ch, DEAD);
				if (status == HCI_ERR_PAGE_TIMEOUT)
					rsp = 0xEEEE;
				else
					rsp = CON_RSP_RESOURCE;
				l2ca_connect_cfm(ch, rsp, 0);
			}
		}
		__l2ca_put(ch);
	}
	btl_write_unlock(&l2cap_chs);
	DBFEXIT;
	return err;
}

/*
   from HCI
*/
int lp_disconnect_ind(hci_con *con)
{
	l2cap_ch	*ch, *next;

	DBFENTER;

	btl_write_lock(&l2cap_chs);
	btl_for_each_next (ch, l2cap_chs, next) {
		l2ca_hold(ch);
		if (ch->con == con) {
			switch (SETSTATE(ch, DEAD)) {
				case CON_OPEN:
					l2ca_disconnect_ind(ch);
					break;
				case CON_CONFIG:
					if (test_and_clear_bit(L2CAP_CONFIG, &ch->flags)) {
						l2ca_config_cfm(ch, CON_RSP_RESOURCE);
						break;
					}
					/* pass through ... */
				case CON_W4_CONRSP:
				case CON_W4_LCONRSP:
					l2ca_connect_cfm(ch, CON_RSP_RESOURCE, 0);
					break;
				default:
			}
		}
		__l2ca_put(ch);
	}
	btl_write_unlock(&l2cap_chs);
	DBFEXIT;
	return 0;
}

/*
  QOS violation indication...

  add it here

*/

int l2cap_proc_read(char *buf, char **start, off_t offset, int len)
{
	int 		count = 0;
	l2cap_ch	*ch;

	DBFENTER;

	count += sprintf(buf+count, "Channels (%d):\n", l2cap_chs.len);
	btl_read_lock(&l2cap_chs);
	btl_for_each (ch, l2cap_chs) {
		count += sprintf(buf+count, "cid: %#x, bda: %s, psm: %d, refcnt: %d, state: %d\n",
				ch->rcid, bda2str(&ch->bda), ch->psm, atomic_read(&ch->refcnt), STATE(ch));
	}
	btl_read_unlock(&l2cap_chs);


	DBFEXIT;
	return count;
}


/*********************************  INITIALIZATION   **************************************/

int init_lpf(void);
void exit_lpf(void);

struct proc_dir_entry	*l2cap_proc;

int __init init_l2cap(void)
{
	int	err = -EINVAL;
	DBFENTER;

	btl_head_init(&l2cap_protos);
	btl_head_init(&l2cap_chs);

	l2cap_proc = create_proc_info_entry("l2cap", 0, proc_affix, l2cap_proc_read);
	if (l2cap_proc == NULL) {
		BTERROR("Unable to register proc fs entry\n");
		goto err1;
	}

	err = init_lpf();
	if (err) {
		BTERROR("Unable to initialize RFCOMM socket layer\n");
		goto err2;
	}
	
	return 0;
err2:
	remove_proc_entry("l2cap", proc_affix);
err1:
	DBFEXIT;
	return err;
}


void exit_l2cap(void)
{
	DBFENTER;

	exit_lpf();
	remove_proc_entry("l2cap", proc_affix);
	
	DBFEXIT;
}

EXPORT_SYMBOL(l2ca_register_protocol);
EXPORT_SYMBOL(l2ca_unregister_protocol);
EXPORT_SYMBOL(l2ca_connect_req);
EXPORT_SYMBOL(l2ca_connect_rsp);
EXPORT_SYMBOL(l2ca_config_req);
EXPORT_SYMBOL(l2ca_config_rsp);
EXPORT_SYMBOL(l2ca_send_data);
EXPORT_SYMBOL(l2ca_create);
EXPORT_SYMBOL(l2ca_put);


/*
  NOTES:

  1. sig_config_rsp
     What's going on if result is not 0???

  2. sig_reject should find a l2cap_con object which sent command
     with id == reqid
     check con->state and do something
     

     
*/
