/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: rfcomm.h,v 1.29 2002/12/13 19:51:56 kds Exp $

   RFCOMM - RFCOMM protocol for Bluetooth

   Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
*/		

#ifndef _RFCOMM_H
#define _RFCOMM_H

#include <affix/bluetooth.h>
#include <affix/l2cap.h>

#define T1_TIMEOUT			(20*HZ)
#define T2_TIMEOUT			(20*HZ)
#define RFCOMM_DEFER_DISC_TIMEOUT	(2*HZ)
#define RFCOMM_CONREQ_TIMEOUT		(1*HZ)
#define RFCOMM_CFGREQ_TIMEOUT		(1*HZ)
#define RFCOMM_CONRSP_TIMEOUT		(60*HZ)	/* pin code requires user interaction */
#define RFCOMM_DISCRSP_TIMEOUT		(2*HZ)


/****************** RFCOMM protocol data types *****************/


/* The values in the control field - frame types */
#define SABM		0x2f
#define SABM_SIZE	4
#define UA		0x63
#define UA_SIZE		4
#define DM		0x0f
#define DISC		0x43
#define UIH		0xef
#define CREDIT_SIZE	5

/* The values in the type field in a multiplexer command packet */
#define TEST		0x08
#define FCON		0x28
#define FCOFF		0x18
#define R_MSC		0x38
#define RPN		0x24
#define RLS		0x14
#define PN		0x20
#define NSC		0x04


#define SHORT_PAYLOAD_SIZE 	0x7F
#define MAX_PAYLOAD_SIZE	0x7FFF
#define RFCOMM_DEFAULT_PAYLOAD_SIZE	0x7F
#define RFCOMM_GUESS_PAYLOAD_SIZE	0x0FFF	/* 4k */

#define SERVER_CHANNEL(dlci)	(dlci >> 1)
#define DIRECTION(dlci)		(dlci & 0x01)
#define DLCI(channel)	(channel << 1)

/* Typedefinitions of stuctures used for creating and parsing packets */
typedef struct address_field {
#if defined(__LITTLE_ENDIAN_BITFIELD)
	__u8 	ea:1;
	__u8 	cr:1;
	__u8	dlci:6;
#elif defined(__BIG_ENDIAN_BITFIELD)
	__u8	dlci:6;
	__u8	cr:1;
	__u8	ea:1;
#else
#error "Bitfield endianness not defined! Check your byteorder.h"
#endif
}__PACK__ address_field;

typedef struct short_length {
#if defined(__LITTLE_ENDIAN_BITFIELD)
	__u8	ea:1;
	__u8	len:7;
#elif defined(__BIG_ENDIAN_BITFIELD)
	__u8	len:7;
	__u8	ea:1;
#endif
}__PACK__ short_length;

#define GET_LENGTH(len)	(__btoh16(len)>>1)
#define SET_LENGTH(len)	(__htob16(len<<1))

typedef struct long_length {
	__u16	len;
}__PACK__ long_length;

typedef struct short_frame {
	address_field addr;
	__u8 control;
	short_length length;
	__u8 data[0];		/* if no data then one byte of FCS */
}__PACK__ short_frame;

typedef struct long_frame {
	address_field addr;
	__u8 control;
	long_length length;
	__u8 data[0];
}__PACK__ long_frame;

typedef struct credit_short_frame {
	address_field addr;
	__u8 control;
	short_length length;
	__u8 credit;
	__u8 data[0];		/* if no data then one byte of FCS */
}__PACK__ credit_short_frame;

typedef struct credit_long_frame {
	address_field addr;
	__u8 control;
	long_length length;
	__u8 credit;
	__u8 data[0];
}__PACK__ credit_long_frame;


#define SET_PF(ctr) ((ctr) | (1 << 4)) 
#define CLR_PF(ctr) ((ctr) & 0xef)
#define GET_PF(ctr) (((ctr) >> 4) & 0x1)

#define UIH_CRC_CHECK	2
#define CTRL_CRC_CHECK	3

#define SHORT_HDR_SIZE	sizeof(short_frame)
#define LONG_HDR_SIZE	sizeof(long_frame)

#define FCS_SIZE	1

/* Typedefinitions for structures used for the multiplexer commands */

#define MCC_CMD		1
#define MCC_RSP		0

typedef struct mcc_type {
#if defined(__LITTLE_ENDIAN_BITFIELD)
	__u8	ea:1;
	__u8	cr:1;
	__u8	value:6;
#elif defined(__BIG_ENDIAN_BITFIELD)
	__u8	value:6;
	__u8	cr:1;
	__u8	ea:1;
#endif
}__PACK__ mcc_type;

typedef struct mcc_frame {
	mcc_type	type;
	short_length	length;
	__u8		value[0];
}__PACK__ mcc_frame;

typedef struct cmd_frame {
	short_frame 	short_pkt;
	mcc_frame 	mcc_pkt;
}__PACK__ cmd_frame;


/* Multiplexer Control Channel Commands */

/* MSC-command */

#define RFCOMM_RTC	0x01
#define RFCOMM_RTR	0x02
#define RFCOMM_IC	0x10
#define RFCOMM_DV	0x20

typedef struct modem_signals {
#if defined(__LITTLE_ENDIAN_BITFIELD)
	__u8	ea:1;
	__u8	fc:1;
	__u8	mr:6;
#elif defined(__BIG_ENDIAN_BITFIELD)
	__u8	mr:6;
	__u8	fc:1;
	__u8	ea:1;
#endif
}__PACK__ modem_signals;

typedef struct break_signals {
#if defined(__LITTLE_ENDIAN_BITFIELD)
	__u8	ea:1;
	__u8	b1:1;
	__u8	b2:1;
	__u8	b3:1;
	__u8	len:4;
#elif defined(__BIG_ENDIAN_BITFIELD)
	__u8	len:4;
	__u8	b3:1;
	__u8	b2:1;
	__u8	b1:1;
	__u8	ea:1;
#endif
}__PACK__ break_signals;

typedef struct msc_msg {
	address_field 		addr;
	modem_signals 		sig;
	//break_signals 		break_signals;
}__PACK__ msc_msg;

typedef struct msc_msg_full {
	address_field 		addr;
	modem_signals 		sig;
	break_signals 		break_signals;
}__PACK__ msc_msg_full;


/* RPN command */
typedef struct parameter_mask {
#if defined(__LITTLE_ENDIAN_BITFIELD)
	__u8	bit_rate:1;
	__u8	data_bits:1;
	__u8	stop_bit:1;
	__u8	parity:1;
	__u8	parity_type:1;
	__u8	xon_char:1;
	__u8	xoff_char:1;
	__u8	res1:1;
	__u8	xon_input:1;	/* next byte */
	__u8	xon_output:1;
	__u8	rtr_input:1;
	__u8	rtr_output:1;
	__u8	rtc_input:1;
	__u8	rtc_output:1;
	__u8	res2:2;
#elif defined(__BIG_ENDIAN_BITFIELD)
	__u8	res1:1;
	__u8	xoff_char:1;
	__u8	xon_char:1;
	__u8	parity_type:1;
	__u8	parity:1;
	__u8	stop_bit:1;
	__u8	data_bits:1;
	__u8	bit_rate:1;
	__u8	res2:2;		/* next byte */
	__u8	rtc_output:1;
	__u8	rtc_input:1;
	__u8	rtr_output:1;
	__u8	rtr_input:1;
	__u8	xon_output:1;
	__u8	xon_input:1;
#endif
}__PACK__ parameter_mask;

typedef struct port_param {
	__u8	bit_rate;
#if defined(__LITTLE_ENDIAN_BITFIELD)
	__u8	data_bits:2;
	__u8	stop_bit:1;
	__u8	parity:1;
	__u8	parity_type:2;
	__u8	res1:2;
	__u8	xon_input:1;	// next byte
	__u8	xon_output:1;
	__u8	rtr_input:1;	/* use CTS */
	__u8	rtr_output:1;	/* use RTS */
	__u8	rtc_input:1;	/* use DSR */
	__u8	rtc_output:1;	/* use DTR */
	__u8	res2:2;
#elif defined(__BIG_ENDIAN_BITFIELD)
	__u8	res1:2;
	__u8	parity_type:2;
	__u8	parity:1;
	__u8	stop_bit:1;
	__u8	data_bits:2;
	__u8	res2:2;	// next byte
	__u8	rtc_output:1;
	__u8	rtc_input:1;
	__u8	rtr_output:1;
	__u8	rtr_input:1;
	__u8	xon_output:1;
	__u8	xon_input:1;
#endif
	__u8 xon_char;
	__u8 xoff_char;
}__PACK__ port_param;

typedef struct rpn_msg {
	address_field		addr;
	port_param		param;
	parameter_mask		pm;
}__PACK__ rpn_msg;

/* RLS-command */  
typedef struct rls_msg {
	address_field 	addr;
#if defined(__LITTLE_ENDIAN_BITFIELD)
	__u8	line_status:4;
	__u8	res:4;
#elif defined(__BIG_ENDIAN_BITFIELD)
	__u8	res:4;
	__u8	line_status:4;
#endif
}__PACK__ rls_msg;

/* PN-command */
typedef struct pn_msg {
#if defined(__LITTLE_ENDIAN_BITFIELD)
	__u8	dlci:6;			/* DLCI */
	__u8	res1:2;
	__u8	unused_frame_type:4;
	__u8	flow:4;			/* Credit based flow control */
	__u8	priority:6;			/* XXX set this */
	__u8	res2:2;
	__u8	unused_ack_timer;
	__u16	frame_size;			/* payload MTU */
	__u8	unused_max_num_retrans;
	__u8	credit:3;			/* number of credits */
	__u8	res3:5;
#elif defined(__BIG_ENDIAN_BITFIELD)
	__u8	res1:2;
	__u8	dlci:6;
	__u8	flow:4;
	__u8	unused_frame_type:4;
	__u8	res2:2;
	__u8	priority:6;
	__u8	unused_ack_timer;
	__u16	frame_size;			/* payload MTU */
	__u8	unused_max_num_retrans;
	__u8	res3:5;			/* number of credits */
	__u8	credit:3;
#endif

}__PACK__ pn_msg;

/* NSC-command */
typedef struct nsc_msg {
	mcc_type cmd_type;
}__PACK__ nsc_msg;


/****************** LOCAL FUNCTION DECLARATION SECTION **********************/

#define STATUS_SUCCESS	0
#define STATUS_FAILURE	1

typedef struct rfcomm_con rfcomm_con;

typedef struct {
	int (*data_ind)(rfcomm_con *con, struct sk_buff *skb);
	int (*connect_ind)(rfcomm_con *con);
	int (*connect_cfm)(rfcomm_con *con, int status);
	int (*disconnect_ind)(rfcomm_con *con);
	int (*control_ind)(rfcomm_con *con, int type);
} rfcomm_proto_ops;

typedef struct {
	struct list_head	q;
	__u8			channel;
	rfcomm_proto_ops	*ops;
} rfcomm_proto;


typedef struct rfcomm_sn {
	struct list_head	list;

	atomic_t		refcnt;
	con_state		state;

	int 			initiator;

	struct timer_list	timer;

	BD_ADDR			bda;
	l2cap_ch		*ch;
//	hci_struct		*hci;

	__u8			hdrlen;

	btlist_head_t		cons;		/* virtual cons */

	/* flow control members */
	__u8			peer_fc;	/* agregatible fc */
	__u8			fc;

	/* credit based flow control */
	int	credit_based;
	int	local_credit;		/* Initial */
	int	remote_credit;		/* Initial */
} rfcomm_sn;


/* control events */

#define RFCOMM_CONTROL_CHANNEL	0x00
#define	RFCOMM_RX_START		0x01
#define RFCOMM_RX_STOP		0x02
#define RFCOMM_BUFFER_REQUEST	0x03

#define RFCOMM_LINE_STATUS	0x10
#define RFCOMM_LINE_STATE	0x11
#define RFCOMM_BREAK_SIGNAL	0x12
#define RFCOMM_TX_START		0x13
#define	RFCOMM_TX_STOP		0x14
#define	RFCOMM_PORT_PARAM	0x15
#define RFCOMM_WRITE_SPACE	0x16


/*  parameters of DTELINE_STATE  --  MCR */
#define BTY_MCR_DTR	RFCOMM_RTC
#define BTY_MCR_RTS	RFCOMM_RTR
#define BTY_MCR_DCD	RFCOMM_DV
#define BTY_MCR_RI	RFCOMM_IC

/*  parameters of DCELINE_STATE  --  MSR */
#define BTY_MSR_CTS	RFCOMM_RTR
#define BTY_MSR_DSR	RFCOMM_RTC
#define BTY_MSR_RI	RFCOMM_IC
#define BTY_MSR_DCD	RFCOMM_DV

/*  parameters of LINE_STATUS */

#define BTY_LSR_BI     0x01    /* Break interrupt indicator */
#define BTY_LSR_OE     0x02    /* Overrun error indicator */
#define BTY_LSR_PE     0x04    /* Parity error indicator */
#define BTY_LSR_FE     0x08    /* Frame error indicator */

/* data format */
#define RFCOMM_CS5	0x00
#define RFCOMM_CS6	0x01
#define RFCOMM_CS7	0x02
#define RFCOMM_CS8	0x03
#define RFCOMM_PARODD	0x00
#define RFCOMM_PAREVEN	0x01
#define RFCOMM_PARMARK	0x02
#define RFCOMM_PARSPC	0x03

typedef volatile enum {
	RFCON_CFG_PN = 1
} rfcon_cfg_state;


struct rfcomm_con {
	struct list_head	list;	/* for queueing purpose */

	atomic_t		refcnt;
	con_state		state;
	rfcon_cfg_state		substate;

	void			*priv;	/* "sock" structure */
	int			line;	/* "tty" line	*/
	
	struct rfcomm_sn	*sn;
	__u8			hdrlen;

	struct timer_list	timer;

	hci_struct		*hci;
	BD_ADDR			bda;
	__u8			dlci;
	rfcomm_proto_ops	*ops;

	/* parametes */
	__u16			peer_frame_size;
	__u16			frame_size;
	__u8			peer_priority;
	__u8			priority;

	/* remote port parameters */
	port_param		peer_param;
	parameter_mask		peer_pm;
	port_param		param;
	parameter_mask		pm;

	/* modem status */
	modem_signals		peer_modem;	/* like MSR */
	modem_signals		modem;		/* like MCR */
	break_signals		peer_break_signal;
	break_signals		break_signal;

	/* line status */
	__u8			peer_line_status;
	__u8			line_status;

	/* credit based flow control stuff*/
	int	credit_based;		/* TRUE if credit based */
	int	overall_credit;		/* local */
	int	local_credit;		/* I can receive */
	int	remote_credit;		/* I can send */

	/* connection type: socket/bty */
	int	flow;
	int	power;
};



/* Function Declarations */
int rfcomm_register_protocol(__u16 *channel, rfcomm_proto_ops *ops, void *param);
void rfcomm_unregister_protocol(__u16 channel);

void rfcon_timer(unsigned long data);
rfcomm_con *rfcon_create(BD_ADDR *bda, __u8 dlci, rfcomm_proto_ops *ops);
void rfcon_destroy(rfcomm_con *con);
void __rfcon_destroy(rfcomm_con *con);
void rfcon_put(rfcomm_con *con);
int rfcon_connect_req(rfcomm_con *con);
int rfcon_connect_rsp(rfcomm_con *con, int status);
int rfcon_disconnect_req(rfcomm_con *con);

void rfcon_set_flow(rfcomm_con *con, int flow);
void rfcon_update_flow(rfcomm_con *con);
void rfcon_set_mcr(rfcomm_con *con, __u8 mcr);
void rfcon_set_param(rfcomm_con *con, port_param *p);
void rfcon_set_break(rfcomm_con *con, __u8 b);


void rfcomm_timer(unsigned long data);
rfcomm_sn *rfcomm_alloc(void);
rfcomm_sn *rfcomm_create(BD_ADDR *bda);
int rfcomm_destroy(rfcomm_sn *sn);
void rfcomm_put(rfcomm_sn *sn);
rfcomm_sn *rfcomm_lookup(BD_ADDR *bda);
int rfcomm_connect(rfcomm_sn *sn);
int rfcomm_connect_req(rfcomm_con *con, rfcomm_sn **ret);
int rfcomm_disconnect_req(rfcomm_sn *sn);
int rfcomm_connect_cfm(rfcomm_sn *sn, int status);
int rfcomm_configure(rfcomm_sn *sn);
int rfcomm_config_complete(rfcomm_sn *sn, int status);

/* commands */
int send_sabm(rfcomm_sn *sn, __u8 dlci);
int send_ua(rfcomm_sn *sn, __u8 dlci);
int send_disc(rfcomm_sn *sn, __u8 dlci);
int send_dm(rfcomm_sn *sn, __u8 dlci);
int _send_dm(rfcomm_sn *sn, __u8 dlci, int solicited);
int send_uih(rfcomm_sn *sn, __u8 dlci, __u8 *data, uint len);

int rfcomm_send_data(rfcomm_con *con, struct sk_buff *skb);
int rfcomm_send_credit(rfcomm_con *con, int credit);
int send_cmd_msc(rfcomm_con *con, __u8 cr);


/* CRC stuff */
void create_crc_table(__u8 table[]);
__u8 crc_calc(__u8 *data, uint length);
__u32 crc_check(__u8 *data, uint length, __u8 check_sum);


int send_cmd_pn(rfcomm_con *con, __u8 cr);

static inline void rfcon_hold(rfcomm_con *con)
{
	atomic_inc(&con->refcnt);
}


static inline void __rfcon_put(rfcomm_con *con)
{
	if (atomic_dec_and_test(&con->refcnt))
		__rfcon_destroy(con);
}

static inline void rfcon_start_timer(rfcomm_con *con, unsigned long timeout)
{
	if (!del_timer(&con->timer))
		rfcon_hold(con);
	mod_timer(&con->timer, jiffies+timeout);
}

static inline void rfcon_stop_timer(rfcomm_con *con)
{
	if (del_timer(&con->timer))
		rfcon_put(con);
}

static inline void rfcomm_hold(rfcomm_sn *sn)
{
	atomic_inc(&sn->refcnt);
}

static inline void rfcomm_start_timer(rfcomm_sn *sn, unsigned long timeout)
{
	if (!del_timer(&sn->timer))
		rfcomm_hold(sn);
	mod_timer(&sn->timer, jiffies+timeout);
}

static inline void rfcomm_stop_timer(rfcomm_sn *sn)
{
	if (del_timer(&sn->timer))
		rfcomm_put(sn);
}



static inline int rfcomm_connected(rfcomm_sn *sn)
{
	if (STATE(sn) == CON_OPEN)
		return 1;

	return 0;
}

static inline int rfcon_ready(rfcomm_con *con)
{
	if (STATE(con) == CON_OPEN)
		return 1;

	return 0;
}

static inline int rfcon_disabled(rfcomm_con *con)
{
	__u8	ret;

	if (con->credit_based)
		return con->remote_credit == 0;
	else {
		ret = con->sn->peer_fc && !con->peer_modem.fc &&
			(!con->peer_param.rtr_output || con->peer_modem.mr & RFCOMM_RTR);
	}

	return !ret;
}

static inline __u8 rfcon_get_msr(rfcomm_con *con)
{
	return con->peer_modem.mr;
}

static inline __u8 rfcon_get_mcr(rfcomm_con *con)
{
	return con->modem.mr;
}


static inline int rfcon_disconnect_pend(rfcomm_con *con)
{
	return (STATE(con) == CON_W4_DISCRSP);
}

static inline __u8 rfcon_get_line_status(rfcomm_con *con)
{
	return con->peer_line_status;
}

static inline __u16 rfcomm_getmtu(rfcomm_sn *sn)
{
	return l2ca_getmtu(sn->ch);
}

/*
  XXX frame_size -->> including header??????
*/
static inline __u16 rfcon_getmtu(rfcomm_con *con)
{
	__u16	rfcomm_mtu = rfcomm_getmtu(con->sn);
	int	mtu;

	/* calculate L2CAP based MTU */
	mtu = rfcomm_mtu - SHORT_HDR_SIZE - FCS_SIZE - con->credit_based;
	if ( mtu > SHORT_PAYLOAD_SIZE)
		mtu--; /* put to big packet */
	mtu = btmin(con->peer_frame_size - con->credit_based, mtu);
	if (mtu < 0)
		mtu = 0;
	return mtu;
}

static inline int rfcon_skb_tailroom(struct sk_buff *skb)
{
	return skb_tailroom(skb)-FCS_SIZE;
}

static inline struct sk_buff *rfcomm_alloc_cmd_skb(rfcomm_sn *sn, int cmdlen)
{
	struct sk_buff	*skb;
	int		hdrlen = sn->hdrlen + sizeof(cmd_frame);

	if (cmdlen+sizeof(mcc_frame) > SHORT_PAYLOAD_SIZE)
		hdrlen++;

	skb = alloc_skb(hdrlen + cmdlen + FCS_SIZE, GFP_ATOMIC);
	if (skb == NULL)
		return NULL;
	skb_reserve(skb, hdrlen);
	skb_put(skb, cmdlen);

	return skb;
}

static inline void rfcon_bind_session(rfcomm_con *con, rfcomm_sn *sn)
{
	if (con->sn)
		rfcomm_put(con->sn);
	rfcomm_hold(sn);
	con->sn = sn;
	con->hdrlen = sn->hdrlen + LONG_HDR_SIZE + 1;	/* XXX: 1 - size of credit field */
	btl_add_tail(&sn->cons, con);
}

static inline void rfcon_setframesize(rfcomm_con *con, int size)
{
	int	n;

	con->frame_size = size;
	for (n=0; size !=0; size>>=1,n++);
	con->power = n;
}

static inline void rfcon_setbuffer(rfcomm_con *con, int size)
{
	/* 
	 * calculates credits from available buffer size
	 */
	if (con->flow == FLOW_ON)
		con->overall_credit = size >> con->power;
	else
		con->overall_credit = 0;

	DBPRT(DBCTRL, "Buffer size: %d, frame size: %d, Overall credit: %d, power: %d\n",
			size, con->frame_size, con->overall_credit, con->power);
}

#endif
