/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Muller Ulrich <ulrich.muller@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


/*
   $Id: pan.c,v 1.36 2002/11/13 00:26:01 kds Exp $

   pan.c - PAN kernel module
   
   Fixes:
   
   	Dmitry Kasatkin		- Integrated to the Affix


   - device registration/unregistration

   This module implements a PAN User, Group Ad-Hoc Network and Network Access Point.
   The role is setup at startup. Functionality:
   PAN User:
   - registers net device "panx"
   - register BNEP layer as upper layer of L2CAP of Bluetooth device "bt0"
   - create connection to GAN/NAP
   - pass all packets from BNEP to net device and vica versa
   Group Ad-Hoc Network/Network Access Point:
   - registers net device "panx"
   - register BNEP layer as upper layer of L2CAP of Bluetooth device "bt0"
   - accept connections from PAN Users
   - route packets between net device and all BNEP connections
   A NAP passes all packets to the net device that are not addressed to a member
   of the Piconet; in contract, a GAN dropps these packets.

   The module gets initialized and closed with the init_module() and cleanup_module()
   fuctions at the bottom. It registeres an upper L2CAP layer (pan_* functions) immediatley.
   The interface to the net device are the bt_* functions. btd_* functions are used to
   manage the connections. Connections are initiated by the connect_thread(). The main
   reason to put the connection initiation into a thread is that it runs in normal context,
   whereas so_* and bt_* functions are usually called in restricted context, e.g. interrupts
   disabled. Thus the connection thread can have more functionality, e.g. call HCI functions.

*/


/* kernel includes: */
#include <linux/config.h>

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/netdevice.h>
#include <linux/proc_fs.h>
#include <linux/if_arp.h>
#include <net/ip.h>
#include <linux/vmalloc.h>
#include <linux/errno.h>
#include <linux/etherdevice.h>
#include <asm/uaccess.h>
#include <linux/smp_lock.h>

/* other includes: */
#define FILEBIT		DBPAN

#include <affix/bluetooth.h>
#include <affix/btdebug.h>
#include <affix/hci.h>
#include <affix/l2cap.h>

#include "pan.h"
#include "bnep.h"


static int errno;

/* global variables */

struct bt_dev		*btdev0;			/* Bluetooth device */

#ifdef CONFIG_PROC_FS
const char		proc_name[] = "pan";		/* proc file system entry name */
struct proc_dir_entry 	*proc_entry;			/* proc file system entry operations */
#endif

#define SigConnectCfm_Pos 1
#define SigConnectCfm_Neg 2
static int SigConnectCfm = 0;				/* set when pos/neg ConnectCfm is received */
static int SigDisconnectCfm = 0;			/* set when DisconnectCfm is received */
static DECLARE_WAIT_QUEUE_HEAD(SigConnectCfm_wait);	/* waked up when ConnectCfm/Ind is received */
static DECLARE_WAIT_QUEUE_HEAD(SigDisconnectCfm_wait);	/* waked up when DisconnectCfm/Ind is received */


/* class of device field according Bluetooth Assigned Numbers (network byte order)*/
__u32	HCI_COD_NETCOMP = 0x020100;
__u32	HCI_COD_LAP_LOAD[8] =	{0x020300, 0x020320, 0x020340, 0x020360,
				0x020380, 0x0203A0, 0x0203C0, 0x0203E0};

/* convert bluetooth CoD to string */

char *BD_CLASS2str(__u32 bdc)
{
	static unsigned char buf[2][9];
	static int num = 0; /* we use function with up to two calls in one printk statement, this requires
			that there are two buffers in memory */

	if (bdc == HCI_COD_NETCOMP)
		return "Computer";
	if (bdc == HCI_COD_LAP_LOAD[0])
		return "LAN Access Point/no connections";
	if (bdc == HCI_COD_LAP_LOAD[1])
		return "LAN Access Point/1 connection";
	if (bdc == HCI_COD_LAP_LOAD[2])
		return "LAN Access Point/2 connections";
	if (bdc == HCI_COD_LAP_LOAD[3])
		return "LAN Access Point/3 connections";
	if (bdc == HCI_COD_LAP_LOAD[4])
		return "LAN Access Point/4 connections";
	if (bdc == HCI_COD_LAP_LOAD[5])
		return "LAN Access Point/5 connections";
	if (bdc == HCI_COD_LAP_LOAD[6])
		return "LAN Access Point/6 connections";
	if (bdc == HCI_COD_LAP_LOAD[7])
		return "LAN Access Point/7 connections";
	num = 1 - num;
	sprintf(buf[num], "%#.6x", bdc);
	return buf[num];
}


/* copy bdaddr to reverse orderd ethernet addr */

void BD_ADDR2ETH_ADDR(void *eta, BD_ADDR *bda)
{
	bdacpy(eta, bda);
}

/* compare BD_ADDR with reverse ordered ethernet addr */

int compBD_ADDR_ETH_ADDR(BD_ADDR *bda, void *etaddr)
{
	BD_ADDR		eth;
	
	bdacpy(&eth, etaddr);
	return !bda_equal(&eth, bda);
}



/* proc entry */

#ifdef CONFIG_PROC_FS

/* report information when user reads /proc/net/pan */
/* this method is only valid for outputs smaller than a page size */
static int proc_read(char *page, char **start, off_t off, int count, int *eof, void *data)
{
#define pprintf(args...)	if(len + 80 < PAGE_SIZE) len += sprintf(page+len, args);
	struct bt_dev *btdev = btdev0; /* currently, only one device supported */
	int i, len = 0;

	if (!btdev)
		return 0;

	/* create ouput */
	/* - role */
	pprintf("role: ");
	if (btdev->my_role == panu_role) {
		pprintf("PAN User\n");
	} else if (btdev->my_role == nap_role) {
		pprintf("Network Access Point\n");
	} else
		pprintf("Group Ad-hoc Network\n");

	/* - local filters */
	if(btdev->pf.count) {
		pprintf("local protocol filter:\n");
		for (i = 0; i < btdev->pf.count; i++)
			pprintf("- 0x%04x - 0x%04x\n", btdev->pf.protocol[i][0], btdev->pf.protocol[i][1]);
	} else {
		pprintf("no local protocol filter\n");
	}
	if(btdev->mf.count) {
		pprintf("local multicast filter:\n");
		for (i = 0; i < btdev->mf.count; i++)
			pprintf("- %s - %s\n", ETH_ADDR2str(&btdev->mf.multicast[i][0]), ETH_ADDR2str(&btdev->mf.multicast[i][1]));
	} else {
		pprintf("no local multicast filter\n");
	}

	/* remote devices */
	if (btdev->connections.len) {
		struct bnep_con *cl;
		pprintf("connections:\n");
		btl_read_lock(&btdev->connections);
		btl_for_each (cl, btdev->connections) {
			struct bnep_layer *bl = cl->priv;
			pprintf("- %s\n", bda2str(&cl->ch->bda));

			if(bl) {
				if(!bl->setup_complete) {
					pprintf("  connection is not setup by BNEP\n");
				} else {
					/* - remote filters */
					if(bl->pf.count) {
						pprintf("  protocol filter:\n");
						for (i = 0; i < bl->pf.count; i++)
							pprintf("  - 0x%04x - 0x%04x\n", bl->pf.protocol[i][0],
									bl->pf.protocol[i][1]);
					} else {
						pprintf("  no protocol filter\n");
					}
					if(bl->mf.count) {
						pprintf("  multicast filter:\n");
						for (i = 0; i < bl->mf.count; i++)
							pprintf("  - %s - %s\n", ETH_ADDR2str(&bl->mf.multicast[i][0]),
									ETH_ADDR2str(&bl->mf.multicast[i][1]));
					} else {
						pprintf("  no multicast filter\n");
					}
				}
			} else {
				pprintf("  connection is not configured by L2CAP\n");
			}
		}
		btl_read_unlock(&btdev->connections);
	} else {
		pprintf("no connections\n");
	}

	//done:
	/* return output */
	if (off >= len) {
		*eof = 1;
		return 0;
	}

	*start = page + off;
	if (count >= len - off) {
		*eof = 1;
	}
	return( btmin(count, len - off) );
}
#endif

/* helper for bt_dev */

int __init bt_init(struct net_device *dev); /* forward declaration */

int pan_init_device(hci_struct *hci, Role my_role)
{
	struct bt_dev	*btdev;
	int		fd;
	
	if (btdev0)
		return -EEXIST;

	btdev = (struct bt_dev *) vmalloc (sizeof(struct bt_dev));
	if (!btdev) {
		BTERROR("out of memory\n");
		return -ENOMEM;
	}
	memset(btdev, 0, sizeof(struct bt_dev));

	btl_head_init(&btdev->connections);
	init_MUTEX(&btdev->thread_sem);
	btdev->hd = hci->hd;
	btdev->my_role = my_role;
	btdev->bdaddr = hci->bda;
	btdev->hci = -1;

	BTINFO("PAN BD_ADDR: %s\n", bda2str(&btdev->bdaddr));

	fd = hci_open_id(hci->hd);
	if (fd < 0) {
		BTERROR("cannot open BT device %d\n", hci->hd);
		goto fail;
	}

	/* set class of device and disable page/inquiry scan until net device is "up" */
	HCI_WriteScanEnable(fd, HCI_SCAN_OFF);
	if (my_role == panu_role)
		HCI_WriteClassOfDevice(fd, HCI_COD_NETCOMP);
	else
		HCI_WriteClassOfDevice(fd, HCI_COD_LAP_LOAD[0]); /* no connections */

	hci_close(fd);
	
	/* create net device "panx" */
	//strcpy(btdev->net_dev.name, "pan%d"); /* kernel will assign a number */
	if (dev_alloc_name(&btdev->net_dev, "pan%d") < 0)
		goto fail;

	btdev->net_dev.init = bt_init;
	btdev->net_dev.priv = (void*) btdev;

	BD_ADDR2ETH_ADDR(&btdev->net_dev.dev_addr, &btdev->bdaddr);

	if (register_netdevice(&btdev->net_dev) != 0) { /* register network device */
		BTERROR("unable to register network device %s\n", btdev->net_dev.name);
		goto fail;
	}
	BTINFO("PAN net device: %s\n", btdev->net_dev.name);

	btdev0 = btdev;
	return 0;

fail:
	vfree(btdev);
	return -1;
}

/* frees memory of btdev and closes hci handle, unregister netdev */
int pan_exit_device(int hd)
{
	struct bt_dev	*btdev = btdev0;
	
	DBFENTER;

	if (!btdev || btdev->hd != hd)
		return -ENODEV;

	btdev0 = NULL;

	if (btdev->net_dev.name) { /* net_dev is valid */
		unregister_netdevice(&btdev->net_dev);
	}

	vfree(btdev);

	DBFEXIT;
	return 0;
}

/* 
 * add new connection to btdev->connections
 * forward connection to bnep layer; when running as a gan/nap, update class of device,
 * returns 0 on success 
 */
int btd_connection_add(struct bt_dev *btdev, l2cap_ch *ch, State state)
{
	struct bnep_con *cl = (struct bnep_con*) vmalloc(sizeof(struct bnep_con));

	DBFENTER;
	if (!cl)
		return -ENOMEM; /* out of memory */
	
	if (btdev->connections.len >= 7) { /* should not be possible */
		DBPRT(DBCTRL, "too many connections!\n");
		vfree(cl);
		return -EBUSY;
	}

	memset(cl, 0, sizeof(struct bnep_con));

	ch->priv = (void*) cl;
	cl->ch = ch;
	cl->btdev = btdev;
	cl->state = state;

	/* add to head of connection list */
	btl_add_tail(&btdev->connections, cl);

	if(state == configured)
		bnep_init(cl);

	DBFEXIT;
	return 0;
}

/* 
 * change state of connection,
 * forward connection to bnep layer
 */
void btd_connection_change_state(l2cap_ch *ch, State state)
{
	struct bnep_con *cl = (struct bnep_con*) ch->priv;

	DBFENTER;
	if(!cl || cl->state == state)
		return;

	cl->state = state;

	/* we dont expect to change state from configured to unconfigured */
	if(state == configured)
		bnep_init(cl);

	DBFEXIT;
}

/* remove connection from btdev->connections
 * forward connection to bnep layer; when running as a gan/nap, update class of device
 */
void btd_connection_remove(l2cap_ch *ch)
{
	struct bnep_con		*cl = (struct bnep_con*) ch->priv;
	struct bt_dev		*btdev = cl->btdev;

	DBFENTER;
	
	if (!cl)
		return;
	ch->priv = NULL;

	bnep_close(cl);
	btl_unlink(&btdev->connections, cl);
	vfree(cl);

	DBFEXIT;
}

/*
 * close all open connections of given device
 */
void btd_close_connections(struct bt_dev *btdev)
{
	struct bnep_con		*cl;
	l2cap_ch		*ch;
	
	while (btdev->connections.len) {
		cl = BTL_NEXT(&btdev->connections);
		ch = cl->ch;

		DBPRT(DBCTRL, "closing connection to %s\n", bda2str(&ch->bda));
		SigDisconnectCfm = 0;

		btd_connection_remove(ch);
		l2ca_put(ch);
	}

	SigDisconnectCfm = 1;
	wake_up_interruptible(&SigDisconnectCfm_wait);

	DBFEXIT;
}

/*
 * this function set the class of device/scan activity according to the current number of connections
 */
void btd_connections_update(struct bt_dev *btdev)
{
	int counter = btdev->connections.len;

	if(counter > 7)
		counter = 7; /* class of device is defined for up to 7 connections */

	if(counter == btdev->connections_counter_old)
		return;

	if(btdev->my_role != panu_role) { /* switch */
		DBPRT(DBCTRL, "setting CoD to load %d\n", counter);
		HCI_WriteClassOfDevice(btdev->hci, HCI_COD_LAP_LOAD[counter]);
		if(counter == 7) /* maximum number of slaves reached */
			HCI_WriteScanEnable(btdev->hci, HCI_SCAN_OFF); /* disable page/inquiry scan */
		else if(btdev->connections_counter_old == 7) /* maximum number of slaves left */
			HCI_WriteScanEnable(btdev->hci, HCI_SCAN_BOTH);
	}

	btdev->connections_counter_old = counter;
}

/* connection initiation */

/* 
 * thread that connects to other devices until thread_stop != 0
 */
int connect_active_thread(void *p)
{
#define Connections_wanted 1 /* a pan user only wants a connection to one gan/nap */
#define Max_Num_Responses 100
#define Max_Connection_Trys 5
	struct bt_dev	*btdev = p;
	INQUIRY_ITEM	inquiry_list[Max_Num_Responses];	/* list of inquiry results */
	int		connect_counter[Max_Num_Responses];	/* list of connection trys to each inquiry device */

	DBFENTER;

	btdev->hci = hci_open_id(btdev->hd);
	if (btdev->hci < 0) {
		BTERROR("cannot open BT device %d\n", btdev->hd);
		btdev->thread_stop = 1;
	}

	up(&btdev->thread_sem); /* started - unlock */

	while (!btdev->thread_stop) { /* loop until thread_stop is set */
		DECLARE_WAITQUEUE(wait, current);
		__u8 Num_Responses_real;
		int i, ret, counter, Num_Responses;

		BTINFO("PAN: inquiry...\n");
		ret = HCI_Inquiry(btdev->hci, 10, Max_Num_Responses, inquiry_list, &Num_Responses_real);

		if (ret != 0) {
			BTERROR("HCI inquiry error\n");
			btdev->thread_stop = 1;
			break;
		}

		/* filter inquiry result with CoD */
		i = 0;
		Num_Responses = Num_Responses_real;
		while (i < Num_Responses) {
			int ok = 0;
			int j;
			__u32	cod = inquiry_list[i].Class_of_Device;
			
			for (j = 0; j < 8; j++)
				if ((cod & HCI_COD_LAP_LOAD[j]) == HCI_COD_LAP_LOAD[j])
					ok = 1;
			if(!ok) {
				DBPRT(DBCTRL, "skipping %s/%s\n", bda2str(&inquiry_list[i].bda), 
						BD_CLASS2str(inquiry_list[i].Class_of_Device));
				memcpy(&inquiry_list[i], &inquiry_list[i + 1], (Num_Responses - i) * sizeof(INQUIRY_ITEM));
				Num_Responses--;
			} else {
				i++;
			}
		}

		BTINFO("PAN: found %d devices (%d total)\n", Num_Responses, Num_Responses_real);
		memset(&connect_counter, 0, sizeof(connect_counter));

		/* now try to connect to up to Connections_wanted devices */
		i = 0;
		counter = 0; /* counts connection trys */
		while (!btdev->thread_stop && (i < Num_Responses ) && 
				(btdev->connections.len < Connections_wanted)) {
			DBPRT(DBCTRL, "connecting to %s/%s (device %d/%d, try %d/%d)\n", 
					bda2str(&inquiry_list[i].bda), 
					BD_CLASS2str(inquiry_list[i].Class_of_Device), i+1, Num_Responses,
					connect_counter[i] + 1, Max_Connection_Trys);

			SigConnectCfm = 0;
			ret = pan_ConnectReq(&inquiry_list[i].bda, BNEPPSM);
			if(ret == 0) { /* now wait until connection is confirmed or cancelled */
				DBPRT(DBCTRL, "waiting for SigConnectCfm\n");
				add_wait_queue(&SigConnectCfm_wait, &wait);
				current->state = TASK_INTERRUPTIBLE;

				while(!btdev->thread_stop && !SigConnectCfm) {
					if (signal_pending(current))
						btdev->thread_stop = 1;

					current->state = TASK_RUNNING;  // unecessary?
					schedule();
				}

				current->state = TASK_RUNNING;
				remove_wait_queue(&SigConnectCfm_wait, &wait);
				DBPRT(DBCTRL, "waiting finished\n");
			}
			btd_connections_update(btdev);

			/* determine next connection partner */
			if(SigConnectCfm == SigConnectCfm_Pos)
				connect_counter[i] = Max_Connection_Trys;
			else
				connect_counter[i]++;

			do {
				i++;
			} while (i < Num_Responses && connect_counter[i] == Max_Connection_Trys);

			if (i == Num_Responses) {
				i = 0;
				while ( i < Num_Responses && connect_counter[i] == Max_Connection_Trys )
					i++;
			}
		}

		/* 
		 * Now we have tried to connect to all devices from last inquiry 
		 * OR have connected to Connections_wanted devices.
		 * Before we start the next inquiry, we wait until all connections are down.
		 */
		DBPRT(DBCTRL, "waiting for all connections to close\n");
		add_wait_queue(&SigDisconnectCfm_wait, &wait);
		set_current_state(TASK_INTERRUPTIBLE);

		while (!btdev->thread_stop && (btdev->connections.len)) {
			if (signal_pending(current)) {
				btdev->thread_stop = 1;
				break;
			}
			schedule();
			btd_connections_update(btdev);
			set_current_state(TASK_INTERRUPTIBLE);
		}

		set_current_state(TASK_RUNNING);
		remove_wait_queue(&SigDisconnectCfm_wait, &wait);
		DBPRT(DBCTRL, "no connections available\n");
	}

	hci_close(btdev->hci);
	
	up(&btdev->thread_sem); /* thread going down */
	DBFEXIT;
	return 0;
}

/*
 * thread that updates CoD/scan activity until thread_stop != 0
 */
int connect_passive_thread(void *p)
{
	struct bt_dev		*btdev = p;
	DECLARE_WAITQUEUE(wait1, current);
	DECLARE_WAITQUEUE(wait2, current);

	DBFENTER;

	btdev->hci = hci_open_id(btdev->hd);
	if (btdev->hci < 0) {
		BTERROR("cannot open BT device %d\n", btdev->hd);
		btdev->thread_stop = 1;
		return -1;
	}

	up(&btdev->thread_sem); /* started: unlock */
	
	/* enable Inquiry Scan + Page Scan and wait for panu to connect */
	HCI_WriteScanEnable(btdev->hci, HCI_SCAN_BOTH);

	add_wait_queue(&SigConnectCfm_wait, &wait1);
	add_wait_queue(&SigDisconnectCfm_wait, &wait2);
	set_current_state(TASK_INTERRUPTIBLE);

	while(!btdev->thread_stop) {	
		if (signal_pending(current)) {
			btdev->thread_stop = 1;
			break;
		}
		schedule();
		btd_connections_update(btdev);
		set_current_state(TASK_INTERRUPTIBLE);
	}

	set_current_state(TASK_RUNNING);
	remove_wait_queue(&SigConnectCfm_wait, &wait1);
	remove_wait_queue(&SigDisconnectCfm_wait, &wait2);

	HCI_WriteScanEnable(btdev->hci, HCI_SCAN_OFF);
	hci_close(btdev->hci);

	up(&btdev->thread_sem); /* thread going down */
	DBFEXIT;
	return 0;
}


/*
 * start + stop panu/access point
 */
void access_point_start(struct bt_dev *btdev)
{
	btdev->thread_stop = 0;
	down(&btdev->thread_sem); /* assume thread is running: set thread running lock */
	if (kernel_thread(connect_passive_thread, btdev, 0) < 0) {
		up(&btdev->thread_sem); /* unlock */
		BTERROR("PAN: unable to start kernel thread\n");
	} else {
		down(&btdev->thread_sem);	/* wait for thread started */
		if (btdev->thread_stop)
			BTERROR("PAN: unable to start kernel thread\n");
	}
}

void panu_start(struct bt_dev *btdev)
{
	btdev->thread_stop = 0;
	down(&btdev->thread_sem); /* assume thread is running: set thread running lock */
	if (kernel_thread(connect_active_thread, btdev, 0) < 0) {
		up(&btdev->thread_sem); /* unlock */
		BTERROR("PAN: unable to start kernel thread\n");
	} else {
		down(&btdev->thread_sem);	/* wait for thread started */
		if (btdev->thread_stop)
			BTERROR("PAN: unable to start kernel thread\n");
	}
}

void access_point_stop(struct bt_dev *btdev)
{
	/* stop thread */
	btdev->thread_stop = 1;				/* signal thread to stop */
	wake_up_interruptible(&SigConnectCfm_wait);	/* wake thread if it is sleeping */
	wake_up_interruptible(&SigDisconnectCfm_wait);	/* wake thread if it is still sleeping */
	down(&btdev->thread_sem);			/* wait for thread to finish */
	up(&btdev->thread_sem);				/* but dont block restart */

	/* close all open connections */
	btd_close_connections(btdev);
}

void panu_stop(struct bt_dev *btdev)
{
	/* stop thread */
	btdev->thread_stop = 1;				/* signal thread to stop */
	wake_up_interruptible(&SigConnectCfm_wait);	/* wake thread if it is sleeping */
	wake_up_interruptible(&SigDisconnectCfm_wait);	/* wake thread if it is still sleeping */
	down(&btdev->thread_sem);			/* wait for thread to finish */
	up(&btdev->thread_sem);				/* but dont block restart */

	/* close all open connections */
	btd_close_connections(btdev);
}


/*
 * net device functions: called from network layer
 */
int bt_open(struct net_device *dev)
{
	struct bt_dev *btdev = (struct bt_dev*) dev->priv;

	DBFENTER;

	MOD_INC_USE_COUNT; /* module is now in use, not removable */

	/* set device flag 'up' */
	netif_start_queue(dev);

	/* prepare bluetooth connection */
	if(btdev->my_role == panu_role)
		panu_start(btdev);
	else
		access_point_start(btdev);

	DBFEXIT;
	return 0;
}

int bt_close(struct net_device *dev)
{
	struct bt_dev *btdev = (struct bt_dev*) dev->priv;

	DBFENTER;

	/* set device flag 'down' */
	netif_stop_queue(dev);

	/* prepare bluetooth connection */
	if(btdev->my_role == panu_role)
		panu_stop(btdev);
	else
		access_point_stop(btdev);

	MOD_DEC_USE_COUNT; /* module can now be removed */

	DBFEXIT;
	return 0;
}

int bt_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
	struct bt_dev *btdev = (struct bt_dev*) dev->priv;

	switch (cmd) {
		case SIOCSFILTERPROTOCOL: 
			{ /* set protocol filter */
				protocol_filter pf;
				if(copy_from_user(&pf, ifr->ifr_ifru.ifru_data, sizeof(pf)))
					return -EFAULT;
				if(pf.count > PROTOCOL_FILTER_MAX)
					return -ENOMEM;
				bnep_set_filter_protocol(btdev, &pf);
				return 0;
			}

		case SIOCSFILTERMULTICAST: 
			{ /* set multicast filter */
				multicast_filter mf;
				if(copy_from_user(&mf, ifr->ifr_ifru.ifru_data, sizeof(mf)))
					return -EFAULT;
				if(mf.count > MULTICAST_FILTER_MAX)
					return -ENOMEM;
				bnep_set_filter_multicast(btdev, &mf);
				return 0;
			}

		case SIOCGFILTERPROTOCOL: 
			/* get protocol filter */
			if(copy_to_user(ifr->ifr_ifru.ifru_data, &btdev->pf, sizeof(protocol_filter)))
				return -EFAULT;
			return 0;

		case SIOCGFILTERMULTICAST: /* get multicast filter */
			if(copy_to_user(ifr->ifr_ifru.ifru_data, &btdev->mf, sizeof(multicast_filter)))
				return -EFAULT;
			return 0;
	}

	return -EOPNOTSUPP;
}

/* 
 * Set or clear the multicast filter for this device
 */
void bt_set_multicast_list(struct net_device *dev)
{
	struct bt_dev *btdev = (struct bt_dev*) dev->priv;
	multicast_filter mf;
	struct dev_mc_list *dmi = dev->mc_list;
	ETH_ADDR *addr;
	int i;

	mf.count = 0;
	if ((dev->flags & IFF_PROMISC) || (dev->flags & IFF_ALLMULTI) || (dev->mc_count > MULTICAST_FILTER_MAX)) {
		DBPRT(DBCTRL, "all multicast mode\n");
		bnep_set_filter_multicast(btdev, &mf);
		return;
	}

	/* Add addresses */
	DBPRT(DBCTRL, "%d multicast addresses:\n", dev->mc_count);
	for (i = 0; i < dev->mc_count; i++) {
		addr = (ETH_ADDR*) dmi->dmi_addr;
		DBPRT(DBCTRL, "- %s\n", ETH_ADDR2str(addr));

		memcpy(&mf.multicast[i][0], addr, 6);
		memcpy(&mf.multicast[i][1], addr, 6); /* start range == stop range */
		dmi   = dmi->next;
	}
	mf.count = dev->mc_count;
	bnep_set_filter_multicast(btdev, &mf);
}

struct net_device_stats *bt_get_stats(struct net_device *dev)
{
	struct bt_dev *btdev = (struct bt_dev*) dev->priv;

	return &btdev->stats;
}

/* receive a packet from upper layer */
int bt_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct bt_dev *btdev = (struct bt_dev*) dev->priv;
	int result = -1, len;

	len = skb->len;
	result = bnep_process_upper(btdev, skb);

	if (result == 0) {
		btdev->stats.tx_packets++;
		btdev->stats.tx_bytes += len;
		btdev->net_dev.trans_start = jiffies;
	} else {
		btdev->stats.tx_dropped++;
	}

	DBFEXIT;
	return 0;
}

/* called from ourself to transmit packet to net device */
void bt_receive(struct sk_buff *skb, struct bt_dev *btdev)
{
	DBFENTER;

	skb->dev = &btdev->net_dev;

	btdev->stats.rx_packets++;
	btdev->stats.rx_bytes += skb->len;
	skb->dev->last_rx = jiffies;

	DBPRT(DBNET, "received %d \n", skb->len);
	netif_rx(skb);   /* Eat it! */

	DBFEXIT;
}

/* initialize virtual network device 'panx' */
int __init bt_init(struct net_device *dev)
{
	struct bt_dev *btdev = (struct bt_dev*) dev->priv;

	dev->open		= bt_open;
	dev->stop		= bt_close;
	dev->do_ioctl		= (void*)bt_ioctl;
	dev->get_stats		= bt_get_stats;
	dev->hard_start_xmit	= bt_xmit;

	if(btdev->my_role == panu_role) /* get multicast filter from net device */
		dev->set_multicast_list = bt_set_multicast_list;

	ether_setup(dev); 	/* register some kernel internal default ethernet functions (net_init.c) */
	dev->set_mac_address 	= NULL;	/* overwrite default ethernet function */

	dev->hard_header_len	= BNEP_hard_header_len + pan_hard_header_len;
	dev->tx_queue_len	= 25; /* maximum number of frames the kernel should queue for the device.
					 Typically this is around 100 for Ethernet */
	dev_init_buffers(dev); // TODO: necessary?

	return 0;
}

/* PAN L2CAP upper protocol layer */

/* connection establishment, configuration and disconnection are done with four signals each,
   according to BT spec.:

   initiator                         responder
   |   A                             |   A
   |   | confirmation      response  |   |
   request |   |                             V   | indication
   |   -------------------------------   |
   V                                     |
   ---------------------------------------

   After a connection is established, it must be configured to change the L2CAP MTU from 672 to
   1691 (BNEP MTU) octets. The configuration only applys to the direction of its request
   (MTU: max. size the initiator of a ConfReq can accept).
   The initiator will start ConfigReq as soon as he receives ConnectCfm. The responder will start
   ConfigReq after ConfigInd. The state of a L2CAP channel is saved in its corresponding
   struct bnep_con. Data may be sent only if state == configured.
   */

/*
 * Upper layers to l2cap
 */
int pan_ConnectReq(BD_ADDR *bdaddr, __u16 psm)
{
	struct bt_dev *btdev = btdev0; /* currently, only one device supported */
	l2cap_ch *ch;

	ch = l2ca_create(psm, &l2cap_ops);
	if (ch == NULL) {
		DBPRT(DBCTRL, "cannot create NOBLIN connection handle\n");
		return -ENOMEM;
	}

	/* channel configuration */
	l2ca_setmtu(ch, PAN_MTU);

	if (btd_connection_add(btdev, ch, unconfigured) != 0)
		goto fail;
	
	return l2ca_connect_req(ch, bdaddr);
fail:
	l2ca_put(ch);
	return -1;
}

int pan_DataWriteBroadcast(__u16 psm, struct sk_buff *skb)
{
	dev_kfree_skb(skb);
	return -EFAULT;
}

int pan_ConnectInd (l2cap_ch *ch)
{
	struct bt_dev *btdev = btdev0; /* currently, only one device supported */

	BTINFO("PAN: connection request from %s\n", bda2str(&ch->bda));

	if(btdev->my_role == panu_role) /* I am not a gan/nap */
		goto fail;

	/* if gan/nap is not "up", refuse connection */
	if ((test_bit(__LINK_STATE_START, &btdev->net_dev.state) == 0)) {
		DBPRT(DBCTRL, "connection refused - net device is not running\n");
		goto fail;
	}

	/* not needed anymore */
	if (btd_connection_add(btdev, ch, unconfigured) != 0)
		goto fail;

	l2ca_hold(ch);
	/* accept connection but do the channel configuration before */
	l2ca_setmtu(ch, PAN_MTU);

	/* accept connection */
	l2ca_connect_rsp(ch, CON_RSP_SUCCESS, 0);
	/* signal connect thread  */
	wake_up_interruptible(&SigConnectCfm_wait); /* waked up when ConnectionEvent is set */
	return 0;

fail:
	l2ca_connect_rsp(ch, CON_RSP_RESOURCE, 0);
	return 0;
}

int pan_ConnectCfm (l2cap_ch *ch, __u16 result, __u16 status)
{
	DBFENTER;

	switch (result) {
		case CON_RSP_SUCCESS:
			BTINFO("PAN: connection to %s successful\n", bda2str(&ch->bda));
			btd_connection_change_state(ch, configured);
			SigConnectCfm = SigConnectCfm_Pos;
			break;

		case CON_RSP_PENDING:
			DBPRT(DBNET, "connection to %s is pending\n", bda2str(&ch->bda));
			return 0;
			break;

		default:
			BTERROR("PAN: connection to %s failed\n", bda2str(&ch->bda));
			SigConnectCfm = SigConnectCfm_Neg;
			btd_connection_remove(ch);
			l2ca_put(ch);
	}

	/* signal connect thread  */
	wake_up_interruptible(&SigConnectCfm_wait); /* waked up when ConnectionEvent is set */
	DBFEXIT;
	return 0;
}

int pan_ConfigInd (l2cap_ch *ch)
{
	DBFENTER;

	if (l2ca_getmtu(ch) < PAN_MTU) { /* wrong MTU */
		DBPRT(DBCTRL, "wrong MTU (%d instead %d), sending negative ConfigRsp\n", l2ca_getmtu(ch), PAN_MTU);
		l2ca_setmtu(ch, PAN_MTU);
		l2ca_config_rsp(ch, CFG_RSP_PARAMETERS);
		return 0;
	}

	DBPRT(DBNET, "sending positive ConfigRsp\n");
	/* the responder of a positive ConfigReq must always return the MTU it will use */
	l2ca_setmtu(ch, PAN_MTU);
	l2ca_config_rsp(ch, CFG_RSP_SUCCESS);

	//btd_connection_change_state(con, configured);

	DBFEXIT;
	return 0;
}

int pan_ConfigCfm (l2cap_ch *ch, __u16 result)
{
	//if(result != CFG_RSP_SUCCESS)

	//btd_connection_change_state(con, configured);

	return 0;
}

int pan_DisconnectInd (l2cap_ch *ch)
{
	DBFENTER;
	BTINFO("PAN: connection to %s closed\n", bda2str(&ch->bda));

	btd_connection_remove(ch);
	l2ca_put(ch);

	/* signal disconnect task that connection is closed, se comment in btd_close_connections */
	DBPRT(DBCTRL, "sending SigDisconnectCfm\n");
	SigDisconnectCfm = 1;
	wake_up_interruptible(&SigDisconnectCfm_wait);

	DBFEXIT;
	return 0;
}

int pan_ReceiveData (l2cap_ch *ch, struct sk_buff *skb)
{
	struct bnep_con *cl = (struct bnep_con*) ch->priv;
	bnep_process_lower(cl, skb);
	return 0;
}

/* Callback functions for L2CAP layer */
l2cap_proto_ops l2cap_ops = {
owner:		THIS_MODULE,
data_ind:	pan_ReceiveData,
connect_ind:	pan_ConnectInd,
connect_cfm:	pan_ConnectCfm,
config_ind:	pan_ConfigInd,
config_cfm:	pan_ConfigCfm,
disconnect_ind:	pan_DisconnectInd
};



/*
 * Device registration/unregistration notifier
 */

static Role my_role = default_role;

int pan_event(struct notifier_block *nb, unsigned long event, void *arg)
{
	hci_struct	*hci = arg;

	switch (event) {
		case HCIDEV_UP:
			pan_init_device(hci, my_role);
			break;

		case HCIDEV_DOWN:
			pan_exit_device(hci->hd);
			break;
		default:
			break;
	}
	return NOTIFY_DONE;
}

struct notifier_block		pan_notifier_block = {
notifier_call:	pan_event,
};


int pan_init(struct init_pan *arg)
{
	hci_struct		*hci = NULL;
	struct net_device	*dev;
	int			err;

	DBFENTER;
	
	/* evaluate "role" */
	if (arg->mode == 0) {
		rtnl_lock();
		err =  pan_exit_device(btdev0->hd);
		rtnl_unlock();
		return err;
	}

	if (AFFIX_PAN_PANU <= arg->mode && arg->mode <= AFFIX_PAN_GN) {
		my_role = arg->mode;
	} else {
		BTERROR("unknown role '%d':\n", arg->mode);
		return -EINVAL;
	}

	/* try to enable existing devices */
	if (arg->name[0] != '\0') {
		dev = dev_get_by_name(arg->name);
		if (dev)
			hci = DEV_HCI(dev);
	} else {
		hci = hci_select();
	}

	if (!hci)
		return -ENODEV;

	rtnl_lock();
	err = pan_init_device(hci, my_role);
	rtnl_unlock();
	hci_put(hci);

	DBFEXIT;
	return err;
}

int pan_ioctl(unsigned int cmd,  unsigned long arg)
{
	int	err;

	switch (cmd) {
		case BTIOC_INIT_PAN:
			err = pan_init((struct init_pan*)arg);
			break;

		default:
			return -ENOIOCTLCMD;
	}

	return err;
}

	
struct affix_pan_operations	pan_ops = {
owner:	THIS_MODULE,
ioctl:	pan_ioctl,
};


/* init + close module */
extern struct proc_dir_entry	*proc_affix;


int init_pan(void)
{
	int	err = -EINVAL;

	DBFENTER;

	/* examine module parameters */
	btdev0 = NULL;

	/* register upper protocol to L2CAP */
	err = l2ca_register_protocol(BNEPPSM, &l2cap_ops);
	if (err) {
		BTERROR("PAN: error registering L2CAP protocol");
		goto err1;
	}

#ifdef CONFIG_PROC_FS
	/* install proc file entry */
	proc_entry = create_proc_entry(proc_name, S_IFREG | S_IRUGO, proc_affix);
	if (proc_entry == NULL)
		goto err2;
	proc_entry->read_proc = proc_read;
	proc_entry->write_proc = NULL;
#endif

	affix_register_notifier(&pan_notifier_block);
	affix_set_pan(&pan_ops);

	DBFEXIT;
	return 0; /* success */

err2:
	l2ca_unregister_protocol(BNEPPSM);
err1:
	BTERROR("init_module failed\n");
	return err; /* error */
}

void exit_pan(void)
{
	DBFENTER;

	/* stop devices first */
	if (btdev0) {
		rtnl_lock();
		pan_exit_device(btdev0->hd);
		rtnl_unlock();
	}

	affix_set_pan(NULL);
	affix_unregister_notifier(&pan_notifier_block);
	
#ifdef CONFIG_PROC_FS
	if (proc_entry) /* unregister proc entry */
		remove_proc_entry(proc_name, proc_net);
#endif
	l2ca_unregister_protocol(BNEPPSM);

	DBFEXIT;
}

#ifdef MODULE
module_init(init_pan);
module_exit(exit_pan);
#endif

#ifdef MODULE

MODULE_AUTHOR("Muller Ulrich");
MODULE_DESCRIPTION("Affix PAN Profile");
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 4, 15)
MODULE_LICENSE("GPL");
#endif

#endif // MODULE

EXPORT_NO_SYMBOLS;

