/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: af_rfcomm.c,v 1.64 2002/12/13 19:51:56 kds Exp $

   AF_RFCOMM - RFCOMM Address family for socket interface

   Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
*/		


/* The following prevents "kernel_version" from being set in this file. */
#define __NO_VERSION__

#include <linux/config.h>
#include <linux/version.h>

/* Module related headers, non-module drivers should not include */
#include <linux/module.h>
#include <linux/init.h>

/* Standard driver includes */
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/slab.h>
#include <linux/timer.h>
#include <linux/spinlock.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/ioctl.h>
#include <linux/proc_fs.h>
#include <linux/notifier.h>
#include <asm/uaccess.h>
#include <asm/bitops.h>

#include <linux/in.h>
#include <linux/skbuff.h>
#include <linux/socket.h>
#include <linux/net.h>
#include <net/sock.h>

#define FILEBIT	DBAFRFCOMM

/* Local Includes */
#include <affix/bluetooth.h>
#include <affix/btdebug.h>
#include <affix/hci.h>
#include <affix/hci_mgr.h>

#include <affix/rfcomm.h>
#include "bty.h"

typedef struct rfcomm_sock	rfcomm_sock;

struct rfcomm_sock{
	struct list_head	q;

	struct sock	*sk;
	BD_ADDR		bda;
	int		channel, sport;
	rfcomm_con	*con;
	int		type;		/* socket type: RFCOMM_SOCKET/RFCOMM_BTY */
	int		security;	/* for authentication purpose */

#define RFCOMM_XMIT_WAKEUP	1
#define RFCOMM_XMIT_BUSY	2
	unsigned long		xmit_flags;
	struct tasklet_struct	tx_task;
};

static inline rfcomm_sock *rpf_get(struct sock *sk)
{
	return (rfcomm_sock*)sk->prot;
}


btlist_head_t			rfcomm_socks;
extern rfcomm_proto_ops 	rpf_proto;

void rpf_tx_task(void *data);
int rpf_xmit_wakeup(struct sock *sk);

/*
 * debuging stuff
 */
static long sock_count = 0;


/*
  Memory management
*/

long skb_count = 0;

/* 
 * Write buffer destructor automatically called from kfree_skb. 
 */
void rpf_write_space(struct sock *sk)
{
	rfcomm_sock	*rsk = rpf_get(sk);
	
	DBFENTER;
	if (sock_writeable(sk))
		bty_control_ind(rsk->con, RFCOMM_WRITE_SPACE);
	DBFEXIT;
}

/*
 * Allocate a skb from the socket's send buffer.
 * It's not waiting !!!
 */
struct sk_buff *rpf_wmalloc(struct sock *sk, unsigned long size, int force, int priority)
{
	rfcomm_sock	*rsk = rpf_get(sk);
	struct sk_buff	*skb;

	DBFENTER;
	DBPRT(DBXMIT, "allocated: %ld requested: %ld, wmem_alloc: %d, sndbuf: %d\n", 
			skb_count, size, atomic_read(&sk->wmem_alloc), sk->sndbuf);
	if (!size)
		return NULL;
	skb = sock_wmalloc(sk, rsk->con->hdrlen + size + FCS_SIZE, force, priority);
	if (!skb)
		return NULL;
	skb_reserve(skb, rsk->con->hdrlen);
	DBFEXIT;
	return skb;
}

/***************************    Connection Management   ********************************/

static inline void rpf_enable_sock(struct sock *sk)
{
	rfcomm_sock	*rsk = rpf_get(sk);

	if (rsk->type == RFCOMM_SOCKET) {
		/* enable transmission */
		rfcon_set_mcr(rsk->con, rfcon_get_mcr(rsk->con) | RFCOMM_RTR | RFCOMM_RTC | RFCOMM_DV);
		rfcon_set_flow(rsk->con, FLOW_ON);
	}
}

struct sock *rpf_lookup_sock(hci_struct *hci, int sch)
{
	rfcomm_sock	*rsk, *zero = NULL;
	struct sock	*sk = NULL;
	
	btl_read_lock(&rfcomm_socks);
	btl_for_each (rsk, rfcomm_socks) {
		if (rsk->sk->state == CON_LISTEN && rsk->channel == sch) {
			if (!zero && bda_zero(&rsk->bda))
				zero = rsk;
			if (bda_equal(&hci->bda, &rsk->bda))
				break;
		}
	}
	if (!rsk)
		rsk = zero;
	if (rsk) {
		sock_hold(rsk->sk);
		sk = rsk->sk;
	}
	btl_read_unlock(&rfcomm_socks);

	return sk;
}

struct sock *rpf_lookup_bda(BD_ADDR *bda)
{
	rfcomm_sock	*rsk;
	struct sock	*sk = NULL;
	
	btl_read_lock(&rfcomm_socks);
	btl_for_each (rsk, rfcomm_socks) {
		if (bda_equal(bda, &rsk->bda)) {
			sock_hold(rsk->sk);
			sk = rsk->sk;
			break;
		}
	}
	btl_read_unlock(&rfcomm_socks);

	return sk;
}

rfcomm_sock *rpf_alloc(int priority)
{
	rfcomm_sock	*rsk;

	rsk = (rfcomm_sock*)kmalloc(sizeof(rfcomm_sock), priority);
	if (rsk == NULL)
		return NULL;
	memset(rsk, 0, sizeof(rfcomm_sock));

	return rsk;
}

void rpf_destruct(struct sock *sk)
{
	rfcomm_sock	*rsk = rpf_get(sk);

	DBFENTER;
	
	kfree(rsk);

	sock_count--;	/* XXX */

	DBFEXIT;
}

void __rpf_destroy(struct sock *sk)
{
	rfcomm_sock	*rsk = rpf_get(sk);
	struct sk_buff	*skb;
	
	DBFENTER;

	__btl_unlink(&rfcomm_socks, rsk);

	if (rsk->con) {
		rsk->con->priv = NULL;
		rfcon_put(xchg(&rsk->con, NULL));
	}

	if (sk->pair) {
		sock_put(sk->pair);
		sk->pair = NULL;
	}
	skb_queue_purge(&sk->write_queue);
	/* 
	   remove all packets and check if threre is a connection
	   in the queue
	*/
	while((skb = skb_dequeue(&sk->receive_queue)) != NULL) {
		if (skb->sk != sk) {
			/* A pending connection */
			__rpf_destroy(skb->sk);
		}	
		kfree_skb(skb);
	}

	sock_put(sk);

	DBFEXIT;
}

void rpf_destroy(struct sock *sk)
{
	btl_write_lock(&rfcomm_socks);
	__rpf_destroy(sk);
	btl_write_unlock(&rfcomm_socks);

}

struct sock *rpf_create(int family, int priority, int zero_it)
{
	struct sock	*sk;
	rfcomm_sock	*rsk;

	DBFENTER;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	sk  = sk_alloc(family, priority, zero_it);
#else
	sk  = sk_alloc(family, priority, zero_it, NULL);
#endif
	if (sk == NULL)
		return NULL;

	rsk = rpf_alloc(priority);
	if (rsk == NULL) {
		sk_free(sk);
		return NULL;
	}

	rsk->sk = sk;
	sk->prot = (void*)rsk;

	tasklet_init(&rsk->tx_task, (void*)rpf_tx_task, (unsigned long) sk);

	/* set attributes */
	rsk->security = HCI_SECURITY_AUTH;	/* default is on */

	btl_add_tail(&rfcomm_socks, rsk);

	sock_count++;	/* XXX */

	DBFEXIT;
	return sk;
}

void rpf_init_data(struct socket *sock, struct sock *sk)
{
	DBFENTER;
	
	sock_init_data(sock, sk);
	sk->destruct = rpf_destruct;
	
	DBFEXIT;
}


int rpf_wait_for_state(struct sock *sk, __u8 state, int noblock)
{
	int	err = 0;
	long	timeo;
	DECLARE_WAITQUEUE(wait, current);

	DBFENTER;

	timeo = sock_sndtimeo(sk, noblock);

	add_wait_queue(sk->sleep, &wait);
	set_current_state(TASK_INTERRUPTIBLE);

	while (sk->state != state) {
		if (!timeo) {
			err = -EAGAIN;
			break;
		}
		if ((err = sock_error(sk))) 
			break;
		if (signal_pending(current)) {
			err = -ERESTARTSYS;
			break;
		}
		timeo = schedule_timeout(timeo);
//		schedule();
		set_current_state(TASK_INTERRUPTIBLE);
	}

	set_current_state(TASK_RUNNING);
	remove_wait_queue(sk->sleep, &wait);

	DBFEXIT;
	return err;
}

int rpf_connect_req(struct sock *sk, struct sockaddr_affix *addr)
{
	rfcomm_sock 	*rsk = rpf_get(sk);
	rfcomm_con	*con;
	int		err;
	hci_struct	*hci;

	DBFENTER;

	if (addr->port < 1 || addr->port > 30)
		return -EINVAL;

	if (!bda_zero(&addr->local)) {
		hci = hci_lookup_bda(&addr->local);	/* find device */
		if (hci == NULL)
			return -ENODEV;
	} else
		hci = NULL;


	// set address
	rsk->bda = addr->bda;
	rsk->channel = addr->port;
	
	con = rfcon_create(&addr->bda, DLCI(addr->port), &rpf_proto);
	if (con == NULL) {
		if (hci)
			hci_put(hci);
		err = -ENOMEM;
		goto exit;
	}

	rsk->con = con;
	con->priv = sk;
	con->hci = hci;

	/* set connection parameters */
	err = rfcon_connect_req(con);

  exit:
	DBFEXIT;
	return err;
}


void rpf_shutdown(struct sock *sk, int status)
{
	if (status) {
		if (status == 0xEEEE)
			sk->err = ETIMEDOUT;
		else
			sk->err = ECONNRESET;
	} 

        sk->shutdown |= RCV_SHUTDOWN;
	sk->state = CON_CLOSED;
	if (!sk->dead) {
                sk->dead = 1;
		sk->state_change(sk);
        }
}

/* 
   RRF protocol subsystem.  Client of RFCOMM. Callback functions
*/
void rpf_data_ready(struct sock *sk, int len)
{
	rfcomm_sock	*rsk = rpf_get(sk);

	DBFENTER;

	bty_data_ind(rsk->con);

	DBFEXIT;
}

void rpf_state_change(struct sock *sk)
{
	rfcomm_sock	*rsk = rpf_get(sk);
	DBFENTER;

	if (sk->dead) {
		bty_unregister_con(rsk->con->line);
		if (sk->socket == NULL)
			rpf_destroy(sk);	/* not owned by "socket" */
	}

	DBFEXIT;
}

void rpf_rfree(struct sk_buff *skb)
{
	struct sock	*sk = skb->sk;
	rfcomm_sock	*rsk = rpf_get(sk);

	atomic_sub(skb->truesize, &sk->rmem_alloc);

	/* update RFCOMM flow */
	if (rsk->con)
		rfcon_update_flow(rsk->con);
}

int rpf_data_ind(rfcomm_con *con, struct sk_buff *skb)
{
	struct sock	*sk = (struct sock*)con->priv;

	DBFENTER;

#if 0
	affix_sock_rx(sk, skb);
#else
	skb->dev = NULL;
	skb_set_owner_r(skb, sk);
	skb->destructor = rpf_rfree;
	skb_queue_tail(&sk->receive_queue, skb);
	if (!sk->dead)
		sk->data_ready(sk,skb->len);
#endif

	DBFEXIT;
	return 0;
}

struct sock *rpf_dup(struct sock *osk)
{
	struct sock	*sk;

	sk = rpf_create(PF_AFFIX, GFP_ATOMIC, 1);
	if (sk == NULL)
		return NULL;

	rpf_init_data(NULL, sk);

	sk->pair = osk;

	sk->protocol = osk->protocol;
	sk->type     = osk->type;
	sk->destruct = osk->destruct;
	sk->sleep    = osk->sleep;

	sk->socket   = osk->socket;
	sk->priority = osk->priority;

	rpf_get(sk)->type = rpf_get(osk)->type;			/* duplicate type */
	rpf_get(sk)->security = rpf_get(osk)->security;		/* duplicate sec_level */

	return sk;
}

int rpf_connect_ind(rfcomm_con *con)
{
	struct sock	*sk, *pair;
	rfcomm_sock	*rsk;
	struct sk_buff	*skb;

	DBFENTER;

	pair = rpf_lookup_sock(con->hci, SERVER_CHANNEL(con->dlci));
	if (pair == NULL) {
		rfcon_connect_rsp(con, CON_RSP_RESOURCE);
		return -1;
	}

	if (pair->state != CON_LISTEN || pair->ack_backlog >= pair->max_ack_backlog) {
		rfcon_connect_rsp(con, CON_RSP_RESOURCE);
		sock_put(pair);
		return -1;
	}

	if (pair->dead) {
		rfcon_connect_rsp(con, CON_RSP_RESOURCE);
		sock_put(pair);
		return -1;
	}

	sk = rpf_dup(pair);
	if (sk == NULL) {
		rfcon_connect_rsp(con, CON_RSP_RESOURCE);
		sock_put(pair);
		return -1;
	}

	pair->ack_backlog++;

	rfcon_hold(con);
	con->priv = sk;
	rsk = rpf_get(sk);
	rsk->con = con;
	rsk->bda = con->bda;
	rsk->channel = SERVER_CHANNEL(con->dlci);

	/* check access rights here */
	//DBPRT(DBCTRL, "sec_mode: %x, sec_level: %x\n", con->hci->sec_mode, rpf_get(pair)->sec_level);
	if ((con->hci->flags & HCI_SECURITY_SERVICE) && 
		(!(con->hci->flags & HCI_SECURITY_AUTH) && (rpf_get(pair)->security & HCI_SECURITY_AUTH))) {
		/* it has to be always ok */
		hci_con	*link = con->sn->ch->con;
		if (!link->btdev->paired) {
			/* start pairing */
			sk->state = CON_AUTH_PENDING;
			lp_auth_req(link);
			return 0;
		}
	}

	/* set connection parameters */
	rfcon_connect_rsp(con, CON_RSP_SUCCESS);	/* Response is here */

	sk->state = CON_ESTABLISHED;
	skb = alloc_skb(0, GFP_ATOMIC); 
	skb_set_owner_r(skb, sk);
	skb_queue_tail(&pair->receive_queue, skb);
	pair->data_ready(pair, skb->len);

	DBFEXIT;
	return 0;
}

int rpf_connect_cfm(rfcomm_con *con, int status)
{
	struct sock	*sk = (struct sock*)con->priv;
	
	DBFENTER;

	if (status) {
		rpf_shutdown(sk, status);
		return -1;
	}

	sk->state = CON_ESTABLISHED;
	if (!sk->dead)
		sk->state_change(sk);

	DBFEXIT;
	return 0;
}


int rpf_disconnect_ind(rfcomm_con *con)
{
	struct sock	*sk = (struct sock*)con->priv;

	DBFENTER;

	rpf_shutdown(sk, 0);

	DBFEXIT;
	return 0;
}

int rpf_control_ind(rfcomm_con *con, int type)
{
	struct sock	*sk = (struct sock*)con->priv;
	int		size;
	
	DBFENTER;
	switch (type) {
		case RFCOMM_BUFFER_REQUEST:
			size =  sk->rcvbuf-atomic_read(&sk->rmem_alloc);
			rfcon_setbuffer(con, (size < 0) ? 0 : size);
			break;
		case RFCOMM_TX_START:
			rpf_xmit_wakeup(sk);
			// fall through...
		default:
			if (con->line != -1)
				bty_control_ind(con, type);
			else {
				/* may be check flow here */
				if (!sk->dead)
					sk->state_change(sk);
			}
			break;
	}
	DBFEXIT;
	return 0;
}

/* Callback functions for RFCOMM layer */
rfcomm_proto_ops rpf_proto = {
	rpf_data_ind,
	rpf_connect_ind,
	rpf_connect_cfm,
	rpf_disconnect_ind,
	rpf_control_ind
};


/******************************  RFCOMM PROTOCOL FAMILY SUBSYSTEM  **************************/

static inline void rpf_wait_on_queue(struct sock *sk)
{
	DECLARE_WAITQUEUE(wait, current);
	DBFENTER;

	add_wait_queue(sk->sleep, &wait);
	set_current_state(TASK_INTERRUPTIBLE);

	if (skb_peek(&sk->receive_queue) == NULL)
		schedule();

	set_current_state(TASK_RUNNING);
	remove_wait_queue(sk->sleep, &wait);

	DBFEXIT;
}

int rpf_release(struct socket *sock)
{
	struct sock	*sk = sock->sk;
	rfcomm_sock	*rsk;

	DBFENTER;

	if (sk == NULL)
		return 0;

        sock->sk   = NULL;      
        sk->socket = NULL;
	
	rsk  = rpf_get(sk);

	if (rsk->con && rsk->con->line != -1) /* btyied */
		goto exit;

	tasklet_disable(&rsk->tx_task);
	tasklet_kill(&rsk->tx_task);

	if (rsk->sport)	{/* server socket */
		int		count = 0;
		rfcomm_sock	*_rsk;
		/* count registered sockets */
		btl_read_lock(&rfcomm_socks);
		btl_for_each (_rsk, rfcomm_socks) {
			if (_rsk->sport == rsk->sport)
				count++;
		}
		btl_read_unlock(&rfcomm_socks);
		if (count == 1) /* we only use it */
			rfcomm_unregister_protocol(rsk->sport);
	}

	sk->state = CON_CLOSED;
	sk->err = 0;
	sk->shutdown |= SEND_SHUTDOWN;
	sk->dead = 1;

	rpf_destroy(sk);

exit:
	MOD_DEC_USE_COUNT;
	DBFEXIT;
	return 0;
}

int rpf_bind(struct socket *sock, struct sockaddr *addr, int addr_len)
{
	struct sock		*sk = sock->sk;
	rfcomm_sock		*rsk = NULL;
	struct sockaddr_affix	*saddr = (void*)addr;
	int			err = -EINVAL, count = 0;

	DBFENTER;

	/* find if already registered */
	btl_read_lock(&rfcomm_socks);
	if (saddr->port) {
		btl_for_each (rsk, rfcomm_socks) {
			if (saddr->port == rsk->sport) {
				count++;
				if (bda_equal(&saddr->bda, &rsk->bda))
					break;
			}
		}
	}
	if (!rsk) {
		err = 0;
		if (!count)
			err = rfcomm_register_protocol(&saddr->port, &rpf_proto, sk);
		if (!err) {
			rsk = rpf_get(sk);
			rsk->sport = saddr->port;
			rsk->channel = saddr->port;
			rsk->bda = saddr->bda;
		}
	}
	btl_read_unlock(&rfcomm_socks);

	DBFEXIT;
	return err;
}

/*
    Connect to remote device
    Wait until connection established
*/
int rpf_connect(struct socket *sock, struct sockaddr *addr, int alen, int flags)
{
	struct sock		*sk = sock->sk;
	struct sockaddr_affix	*saddr = (void*)addr;
	int			err = 0;

	DBFENTER;

	/* deal with restarts */
	if (sk->state == CON_ESTABLISHED && sock->state == SS_CONNECTING) {
		sock->state = SS_CONNECTED;
		return 0;
	}
		
	if (sk->state == CON_CLOSED && sock->state == SS_CONNECTING) {
		sock->state = SS_UNCONNECTED;
		return -ECONNREFUSED;
	}
	
	if (sk->state == CON_ESTABLISHED)
		return -EISCONN;      /* No reconnect on a seqpacket socket */
	
	/* Move to connecting socket, start sending Connect Requests */
	sock->state = SS_CONNECTING;
	sk->state   = CON_CONNECTING;

	err = rpf_connect_req(sk, saddr);
	if (err < 0)
		goto exit;

	err = rpf_wait_for_state(sk, CON_ESTABLISHED, flags & O_NONBLOCK);
	if (err)
		goto exit;

	if (sk->state == CON_ESTABLISHED) {
		sock->state = SS_CONNECTED;
		rpf_enable_sock(sk);
	}

 exit:
	DBFEXIT;
	return err;
}

/*
  ioctl helpers
*/
int rpf_open_port(struct sock *sk, int *uline)
{
	rfcomm_sock	*rsk = rpf_get(sk);
	int		err, line;
	
	DBFENTER;
	if (sk->state != CON_ESTABLISHED)
		return -ENOTCONN;
	if (get_user(line, uline))
		return -EFAULT;
	err = bty_register_con(rsk->con, line);
	if (err < 0)
		return err;
	/*
	  set some sk fields
	*/
	sk->state_change =  rpf_state_change;
	sk->data_ready = rpf_data_ready;
	sk->write_space = rpf_write_space;
	line = err;
	if (put_user(line, uline))
		return -EFAULT;
	DBFEXIT;
	return 0;
}

int rpf_close_port(int *uline)
{
	int		line;
	rfcomm_con	*con;
	struct sock	*sk;
	
	DBFENTER;

	if (get_user(line, uline))
		return -EFAULT;

	con = bty_getcon(line);
	if (con == NULL)
		return -ENODEV;
	bty_unregister_con(line);
	
	sk = con->priv;
	if (sk == NULL)
		return -ENODEV;

	if (sk->socket)	/* still owned by "socket" */
		return -EBUSY;

	sk->state = CON_CLOSED;
        sk->err = 0;
	sk->shutdown |= SEND_SHUTDOWN;
	sk->dead = 1;

	rpf_destroy(sk);

	DBFEXIT;
	return 0;
}

int rpf_checkports(struct rfcomm_ports *info)
{
	int		line, size, count;
	rfcomm_con	*con;
	struct rfcomm_port	*pi;
	
	DBFENTER;

	if (get_user(size, &info->size))
		return -EFAULT;

	if (get_user(pi, &info->info))
		return -EFAULT;

	// check space for fill
	if (!access_ok(VERIFY_WRITE, pi, size * sizeof(struct rfcomm_port)))
		return -EFAULT;

	DBPRT(DBINIT, "Size: %d\n", size);

 	for (line = 0, count = 0; line < bty_maxdev && count < size; line++) {
		con = bty_getcon(line);
		if (con) {
			__put_user(con->line, &pi->line);
			__copy_to_user(&pi->bda, &con->bda, 6); 
			__put_user(SERVER_CHANNEL(con->dlci), &pi->channel);
			pi++;
			count++;
		}
	}
	
	if (put_user(count, &info->count))
		return -EFAULT;

	DBPRT(DBINIT, "Count: %d\n", count);
	DBFEXIT;
	return 0;
}

int rpf_settype(struct sock *sk, int *utype)
{
	rfcomm_sock	*rsk = rpf_get(sk);
	int		type;

	DBFENTER;

	if (get_user(type, utype))
		return -EFAULT;

	if (type  > RFCOMM_BTY)
		return -EINVAL;

	rsk->type = type;

	DBFEXIT;
	return 0;
}

int rpf_ioctl(struct socket *sock, unsigned int cmd,  unsigned long arg)
{
	int	err = 0;

	DBFENTER;

	switch(cmd) {
		case SIOCRFCOMM_BINDPORT:
			err = rpf_open_port(sock->sk, (int*)arg);
			break;
		case SIOCRFCOMM_DISCONNECTPORT:
			err = rpf_close_port((int*)arg);
			break;
		case SIOCRFCOMM_CHECKPORTS:
			err = rpf_checkports((void*)arg);
			break;
		case SIOCRFCOMM_SETTYPE:
			err = rpf_settype(sock->sk, (int*)arg);
			break;
		default:
			DBPRT(DBCTRL, "Unprocessed ioctl: %x\n", cmd);
			return -ENOIOCTLCMD;
	}

	DBFEXIT;
	return err;
}

/*
   register protocol and allow clients to establish connection
*/
int rpf_listen(struct socket *sock, int backlog)
{
	struct sock	*sk = sock->sk;
	rfcomm_sock	*rsk = rpf_get(sk);

	DBFENTER;

	if (sock->state != SS_UNCONNECTED)
		return -EINVAL;

	if (rsk->sport == 0)
		return -EOPNOTSUPP;	

	if (sk->state != CON_LISTEN) {
		sk->max_ack_backlog = backlog;
		sk->state = CON_LISTEN;
		sock->flags |= __SO_ACCEPTCON;
	}
	
	DBFEXIT;
	return 0;
}


/*
   Accept connection from remote device
   Wait until ready connection will become available
*/

int rpf_accept(struct socket *sock, struct socket *newsock, int flags)
{
	struct sock	*sk = sock->sk, *newsk;
	struct sk_buff	*skb;
	int		err = -EINVAL;

	DBFENTER;

	if (sock->state != SS_UNCONNECTED || !(sock->flags & __SO_ACCEPTCON))
		goto exit;
	
	while( (skb = skb_dequeue(&sk->receive_queue)) == NULL) {
		if (flags & O_NONBLOCK)
			return -EWOULDBLOCK;

		rpf_wait_on_queue(sk);
		if (signal_pending(current)) 
			return -ERESTARTSYS;
	}

	if (newsock->sk) {
		sk_free(newsock->sk);
		newsock->sk = NULL;
	}

	newsk = skb->sk;
	sock_put(newsk->pair);		/* put pair */
	newsk->pair = NULL;
	sock_graft(newsk, newsock);
	sk->ack_backlog--;
	newsock->state = SS_CONNECTED;
	rpf_enable_sock(newsk);
	kfree_skb(skb);
	MOD_INC_USE_COUNT;
	err = 0;
 exit:			
	DBFEXIT;
	return err;
}

int rpf_getname(struct socket *sock, struct sockaddr *uaddr, int *alen, int peer)
{
	struct sock		*sk = sock->sk;
	struct sockaddr_affix	*addr = (struct sockaddr_affix*)uaddr;
	rfcomm_sock		*rsk = rpf_get(sk);

	DBFENTER;

	addr->family = AF_AFFIX;
	addr->bda = rsk->bda;
	addr->port = rsk->channel;
	*alen = sizeof(struct sockaddr_affix);

	DBFEXIT;
	return 0;

}

int rpf_setsockopt(struct socket *sock, int level, int optname, char *optval, int optlen)
{
	rfcomm_sock	*rsk = rpf_get(sock->sk);

	DBFENTER;
	switch (optname) {
		case BTSO_SECURITY:
			if (get_user(rsk->security, (int*)optval))
				return -EFAULT;
			break;
		default:
			return -EINVAL;
	}
	DBFEXIT;
	return 0;
}

int rpf_getsockopt(struct socket *sock, int level, int optname, char *optval, int *optlen)
{
	rfcomm_sock	*rsk = rpf_get(sock->sk);
	int		err = 0;

	DBFENTER;
	switch (optname) {
		case BTSO_MTU:{
			int	mtu;
			if (sock->sk->state != CON_ESTABLISHED)
				return -ENOTCONN;
			mtu = rfcon_getmtu(rsk->con);
			if (put_user(mtu, (int*)optval))
				return -EFAULT;
			if (put_user(sizeof(int), optlen))
				return -EFAULT;
		}
		break;

		default:
			return -EINVAL;
	}
	DBFEXIT;
	return err;
}


int rpf_wait_flow(struct sock *sk, int noblock)
{
	rfcomm_sock	*rsk = rpf_get(sk);
	int		err = 0;
	long	timeo;
	DECLARE_WAITQUEUE(wait, current);

	DBFENTER;
	timeo = sock_sndtimeo(sk, noblock);
	add_wait_queue(sk->sleep, &wait);	/* _exclusive */
	set_current_state(TASK_INTERRUPTIBLE);
	while (rfcon_disabled(rsk->con)) {
		if (!timeo) {
			err = -EAGAIN;
			break;
		}
		if ((err = sock_error(sk))) 
			break;
		/* Socket shut down? */
		if (sk->shutdown & SHUTDOWN_MASK)
			break;
		if (signal_pending(current)) {
			err = -ERESTARTSYS;
			break;
		}
		timeo = schedule_timeout(timeo);
		set_current_state(TASK_INTERRUPTIBLE);
	}
	set_current_state(TASK_RUNNING);
	remove_wait_queue(sk->sleep, &wait);
	DBFEXIT;
	return err;
}

int rpf_xmit_wakeup(struct sock *sk)
{
	rfcomm_sock	*rsk = rpf_get(sk);
	int		err;
	struct sk_buff	*skb;
	
	DBFENTER;
	set_bit(RFCOMM_XMIT_WAKEUP, &rsk->xmit_flags);
restart:
	if (test_and_set_bit(RFCOMM_XMIT_BUSY, &rsk->xmit_flags))
		return 0;
	clear_bit(RFCOMM_XMIT_WAKEUP, &rsk->xmit_flags);
	err = 0;
	for (;;) {
		if (sk->state != CON_ESTABLISHED) {
			err = -ENOTCONN;
			break;
		}
		if (rfcon_disabled(rsk->con))
			break;
		skb = skb_dequeue(&sk->write_queue);
		if (!skb)
			break;
		DBPRT(DBXMIT, "Message is sent, len: %d\n", skb->len);
		DBPRTPKT(DBXMIT, skb->data, skb->len);
		err = rfcomm_send_data(rsk->con, skb);
		if (err)
			break;
	}
	clear_bit(RFCOMM_XMIT_BUSY, &rsk->xmit_flags);
	if (test_bit(RFCOMM_XMIT_WAKEUP, &rsk->xmit_flags))
		goto restart;
	DBFEXIT;
	return err;
}

void rpf_tx_task(void *data)
{
	struct sock	*sk = data;
	DBFENTER;
	rpf_xmit_wakeup(sk);
	DBFEXIT;
}

/*
 * exported for BTY
 *
 * transmit now if now write space
 * otherwise later
 */
void rpf_send_data(struct sock *sk)
{
	rfcomm_sock	*rsk = rpf_get(sk);
	unsigned long	flags;
	struct sk_buff	*skb;
	int		tx_now = 0;
	
	spin_lock_irqsave(&sk->write_queue.lock, flags);
	skb = skb_peek_tail(&sk->write_queue);
	if (skb && rfcon_skb_tailroom(skb) == 0)
		tx_now = 1;
	spin_unlock_irqrestore(&sk->write_queue.lock, flags);
	if (tx_now)
		rpf_xmit_wakeup(sk);
	else
		tasklet_schedule(&rsk->tx_task);
}


/*
    size - size of the iovec array, size of the message
*/
int rpf_sendmsg(struct socket *sock, struct msghdr *msg, int size, struct scm_cookie *scm)
{
	struct sock		*sk = sock->sk;
	rfcomm_sock		*rsk = rpf_get(sk);
	int			err = -ENOTCONN;
	struct sk_buff		*txbuff;
	int			mtu, count = 0;

	DBFENTER;

	if (msg->msg_flags & ~MSG_DONTWAIT)
		return -EINVAL;

	while (size) {
		if (sock->state != SS_CONNECTED) 
			return -ENOTCONN;
		if (sk->shutdown & SHUTDOWN_MASK) {
			if (!(msg->msg_flags & MSG_NOSIGNAL))
				send_sig(SIGPIPE, current, 0);
			return -EPIPE;
		}
		if (sk->state != CON_ESTABLISHED)
			return -ENOTCONN;

		txbuff = skb_dequeue_tail(&sk->write_queue);
		if (txbuff && rfcon_skb_tailroom(txbuff) == 0) {
			/* no space */
			skb_queue_tail(&sk->write_queue, txbuff);
			txbuff = NULL;
		}
		if (!txbuff) {
			/* allocate new one */
			mtu = rfcon_getmtu(rsk->con);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 4)
			txbuff = sock_alloc_send_skb(sk, rsk->con->hdrlen + mtu + FCS_SIZE, 0,
					msg->msg_flags & MSG_DONTWAIT, &err);
#else
			txbuff = sock_alloc_send_skb(sk, rsk->con->hdrlen + mtu + FCS_SIZE,
					msg->msg_flags & MSG_DONTWAIT, &err);
#endif
			if (!txbuff)
				break;
			skb_reserve(txbuff, rsk->con->hdrlen);
		}
		mtu = btmin(size, rfcon_skb_tailroom(txbuff));
		err = memcpy_fromiovec(skb_put(txbuff, mtu), msg->msg_iov, mtu);
		if (err)
			return err;
		count += mtu;
		size -= mtu;
		skb_queue_tail(&sk->write_queue, txbuff);
		rpf_xmit_wakeup(sk);
	}
	DBFEXIT;
	return (!err) ? count : err;
}

int rpf_recvmsg(struct socket *sock, struct msghdr *msg, int size, int flags, struct scm_cookie *scm)
{
	struct sock		*sk = sock->sk;
	int			err = -ENOTCONN;
	struct sk_buff		*skb;
	int			copied;

	DBFENTER;

	if (sock->flags & __SO_ACCEPTCON)
		return -EINVAL;

	if (sock->state != SS_CONNECTED)
		return -ENOTCONN;

	if (sk->state != CON_ESTABLISHED) {
		if (sk->shutdown & SHUTDOWN_MASK) {
			//if (!(flags & MSG_NOSIGNAL))
			//	send_sig(SIGPIPE, current, 0);
			return sock_error(sk);
		}
		return -ENOTCONN;
	}

	skb = skb_recv_datagram(sk, (flags & ~MSG_DONTWAIT) | MSG_PEEK, flags & MSG_DONTWAIT, &err);
	if (skb == NULL)
		goto exit;

	DBPRT(DBRECV, "Message received, len: %d\n", skb->len);
	DBPRTPKT(DBRECV, skb->data, skb->len);

	skb->h.raw = skb->data;

	copied = size < skb->len ? size : skb->len;
	skb_copy_datagram_iovec(skb, 0, msg->msg_iov, copied);
	if (!(flags & MSG_PEEK))
		skb_pull(skb, copied);
	if (!skb->len) {
		skb_unlink(skb);
		kfree_skb(skb);
	}
	kfree_skb(skb);		// remove MSG_PEEK link
	err = copied;
 exit:
	DBFEXIT;
	return err;
}

/*
    Protocol family operations
*/

struct proto_ops rpf_ops = {
	family:		PF_AFFIX,

	release:	rpf_release,
	bind:		rpf_bind,
	connect:	rpf_connect,
	socketpair:	sock_no_socketpair,
	accept:		rpf_accept,
	getname:	rpf_getname,
	poll:		datagram_poll,
	ioctl:		rpf_ioctl,
	listen:		rpf_listen,
	shutdown:	sock_no_shutdown,	/* it is used in ax25 */
	setsockopt:	rpf_setsockopt,
	getsockopt:	sock_no_getsockopt,
	sendmsg:	rpf_sendmsg,
	recvmsg:	rpf_recvmsg,
	mmap:		sock_no_mmap
};

int __rpf_create(struct socket *sock, int protocol)
{
	struct sock		*sk;

	DBFENTER;

	sock->state = SS_UNCONNECTED;

	if (sock->type != SOCK_STREAM)
		return -ESOCKTNOSUPPORT;

	sk = rpf_create(PF_AFFIX, GFP_KERNEL, 1);
	if (sk == NULL)
		return -ENOMEM;

	rpf_init_data(sock, sk);	/* generic initialization */
	sk->protocol = protocol;
	sock->ops = &rpf_ops;

	MOD_INC_USE_COUNT;

	DBFEXIT;
	return 0;
}


struct net_proto_family rpf_family_ops = {
	family:	PF_AFFIX,
	create:	__rpf_create
};


int rpf_proc_read(char *buf, char **start, off_t offset, int len)
{
	int count = 0;

	DBFENTER;

	count += sprintf(buf+count, "skb allocated: %ld\n", skb_count);
	count += sprintf(buf+count, "RFCOMM sockets in use: %ld, list len: %d\n",
			sock_count, rfcomm_socks.len);

	DBFEXIT;
	return count;
}

static inline int rpf_auth_complete(hci_con *link)
{
	struct sock	*sk;
	struct sk_buff	*skb;
	rfcomm_sock	*rsk, *next;
	hci_con		*__link;

	DBFENTER;
	
	btl_read_lock(&rfcomm_socks);
	btl_for_each_next (rsk, rfcomm_socks, next) {
		sk = rsk->sk;
		if (sk->state != CON_AUTH_PENDING || !rsk->con)
			continue;
		__link = rsk->con->sn->ch->con;
		if (__link != link)
			continue;
		if (!link->btdev->paired) {
			rfcon_connect_rsp(rsk->con, CON_RSP_SECURITY);	/* Response is here */
			sk->pair->ack_backlog--;
			__rpf_destroy(sk);
			continue;
		}
		/* set connection parameters */
		rfcon_connect_rsp(rsk->con, CON_RSP_SUCCESS);	/* Response is here */

		sk->state = CON_ESTABLISHED;
		skb = alloc_skb(0, GFP_ATOMIC); 
		skb_set_owner_r(skb, sk);
		skb_queue_tail(&sk->pair->receive_queue, skb);
		sk->pair->data_ready(sk->pair, skb->len);

	}
	btl_read_unlock(&rfcomm_socks);

	DBFEXIT;
	return 0;
}

int affix_event(struct notifier_block *nb, unsigned long event, void *arg)
{
	switch (event) {
		case HCICON_AUTH_COMPLETE:
			rpf_auth_complete(arg);
			break;
		default:
			break;
	}
	return NOTIFY_DONE;
}

struct notifier_block affix_notifier_block = {
	notifier_call:	affix_event,
};

struct proc_dir_entry	*rpf_proc;

int init_rpf(void)
{
	int	err;
	
	DBFENTER;

	rpf_proc = create_proc_info_entry("rfcomm_sock", 0, proc_affix, rpf_proc_read);
	if (rpf_proc == NULL) {
		BTERROR("Unable to register proc fs entry\n");
		return -EINVAL;
	}
	err = affix_sock_register(&rpf_family_ops, BTPROTO_RFCOMM);
	if (err) {
		goto err1;
	}
	affix_register_notifier(&affix_notifier_block);
	btl_head_init(&rfcomm_socks);
	
	DBFEXIT;
	return 0;
err1:
	remove_proc_entry("rfcomm_sock", proc_affix);
	return err;
}

void exit_rpf(void)
{
	DBFENTER;

	affix_unregister_notifier(&affix_notifier_block);
	affix_sock_unregister(BTPROTO_RFCOMM);
	remove_proc_entry("rfcomm_sock", proc_affix);

	DBFEXIT;
}

/*
  Notes:

  - in connect_ind: what if pair->dead == 1. ???

*/
